# cython: language_level=3
# cython: boundscheck=False
# cython: cdivision=True
# cython: wraparound=False
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause
import numpy as np
cimport numpy as np

from libc.stdlib cimport free, malloc
from libc.math cimport INFINITY, floor, log2, pow
from libc.string cimport memset
from numpy cimport uint32_t

from ..distance._cdistance cimport (
    Subsequence,
    SubsequenceMetric,
    SubsequenceMetricList,
    SubsequenceView,
    Metric,
    dilated_distance_profile,
    scaled_dilated_distance_profile,
)
from ..utils cimport TSArray
from ..utils._misc cimport argsort, to_ndarray_int
from ..utils._stats cimport fast_mean_std
from ..utils._rand cimport (
    RAND_R_MAX,
    VoseRand,
    rand_int,
    rand_uniform,
    vose_rand_free,
    vose_rand_init,
    vose_rand_int,
    vose_rand_precompute,
)
from ._attr_gen cimport Attribute, AttributeGenerator


cdef class ShapeletAttributeGenerator(AttributeGenerator):

    cdef Py_ssize_t min_shapelet_size
    cdef Py_ssize_t max_shapelet_size

    cdef readonly SubsequenceMetric metric

    def __init__(self, metric, min_shapelet_size, max_shapelet_size):
        self.metric = metric
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size

    cdef int reset(self, TSArray X) noexcept nogil:
        self.metric.reset(X)
        return 1

    cdef Py_ssize_t free_transient(self, Attribute *attribute) noexcept nogil:
        if attribute.attribute != NULL:
            self.metric.free_transient(<SubsequenceView*> attribute.attribute)
            free(attribute.attribute)

    cdef Py_ssize_t free_persistent(self, Attribute *attribute) noexcept nogil:
        cdef Subsequence *s
        if attribute.attribute != NULL:
            self.metric.free_persistent(<Subsequence*> attribute.attribute)
            free(attribute.attribute)

    cdef Py_ssize_t init_persistent(
        self,
        TSArray X,
        Attribute *transient,
        Attribute *persistent
    ) noexcept nogil:
        cdef SubsequenceView *v = <SubsequenceView*> transient.attribute
        cdef Subsequence *s = <Subsequence*> malloc(sizeof(Subsequence))
        self.metric.init_persistent(X, v, s)
        persistent.dim = transient.dim
        persistent.attribute = s
        return 1

    cdef double transient_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil:
        return self.metric.transient_distance(
            <SubsequenceView*> attribute.attribute, X, sample
        )

    cdef double persistent_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil:
        return self.metric.persistent_distance(
            <Subsequence*> attribute.attribute, X, sample
        )

    cdef Py_ssize_t transient_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t attribute_id,
    ) noexcept nogil:
        out[out_sample, attribute_id] = self.metric.transient_distance(
            <SubsequenceView*> attribute.attribute, X, sample
        )
        return 0

    cdef Py_ssize_t persistent_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t attribute_id,
    ) noexcept nogil:
        out[out_sample, attribute_id] = self.metric.persistent_distance(
            <Subsequence*> attribute.attribute, X, sample
        )
        return 0

    cdef object persistent_to_object(self, Attribute *attribute):
        return attribute.dim, self.metric.to_array(<Subsequence*>attribute.attribute)

    cdef Py_ssize_t persistent_from_object(self, object object, Attribute *attribute):
        cdef Subsequence *s = <Subsequence*> malloc(sizeof(Subsequence))
        dim, obj = object
        self.metric.from_array(s, obj)
        attribute.dim = dim
        attribute.attribute = s
        return 0

cdef class RandomShapeletAttributeGenerator(ShapeletAttributeGenerator):

    cdef Py_ssize_t n_shapelets

    def __init__(
        self, metric, min_shapelet_size, max_shapelet_size, n_shapelets
    ):
        super().__init__(metric, min_shapelet_size, max_shapelet_size)
        self.n_shapelets = n_shapelets

    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil:
        return self.n_shapelets

    cdef Py_ssize_t next_attribute(
        self,
        Py_ssize_t attribute_id,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Attribute *transient,
        uint32_t *random_seed
    ) noexcept nogil:
        if attribute_id >= self.n_shapelets:
            return -1

        cdef Py_ssize_t shapelet_length
        cdef Py_ssize_t shapelet_start
        cdef Py_ssize_t shapelet_index
        cdef Py_ssize_t shapelet_dim
        cdef SubsequenceView *v = <SubsequenceView*> malloc(sizeof(SubsequenceView))

        shapelet_length = rand_int(
            self.min_shapelet_size, self.max_shapelet_size, random_seed)
        shapelet_start = rand_int(
            0, X.shape[2] - shapelet_length, random_seed)
        shapelet_index = samples[rand_int(0, n_samples, random_seed)]
        if X.shape[1] > 1:
            shapelet_dim = rand_int(0, X.shape[1], random_seed)
        else:
            shapelet_dim = 0

        transient.dim = shapelet_dim
        self.metric.init_transient(
            X,
            v,
            shapelet_index,
            shapelet_start,
            shapelet_length,
            shapelet_dim,
        )
        transient.attribute = v
        return 1

cdef struct MetricSubsequenceView:
    Py_ssize_t metric
    SubsequenceView view

cdef struct MetricSubsequence:
    Py_ssize_t metric
    Subsequence subsequence


cdef class MultiMetricShapeletAttributeGenerator(AttributeGenerator):
    cdef Py_ssize_t min_shapelet_size
    cdef Py_ssize_t max_shapelet_size
    cdef SubsequenceMetricList metrics
    cdef bint weighted
    cdef VoseRand vr
    cdef const double[::1] weights

    def __init__(
        self,
        Py_ssize_t min_shapelet_size,
        Py_ssize_t max_shapelet_size,
        list metrics,
        const double[::1] weights=None,
    ):
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metrics = SubsequenceMetricList(metrics)
        self.weights = weights
        if weights is not None:
            vose_rand_init(&self.vr, weights.shape[0])
            vose_rand_precompute(&self.vr, &weights[0])
            self.weighted = True
        else:
            self.weighted = False

    def __dealloc__(self):
        if self.weighted:
            vose_rand_free(&self.vr)

    cdef int reset(self, TSArray X) noexcept nogil:
        cdef Py_ssize_t metric
        for metric in range(self.metrics.size):
            self.metrics.reset(metric, X)

        return 1

    cdef Py_ssize_t free_transient(self, Attribute *attribute) noexcept nogil:
        cdef MetricSubsequenceView *msv
        if attribute.attribute != NULL:
            msv = <MetricSubsequenceView*> attribute.attribute
            self.metrics.free_transient(msv.metric, &msv.view)
            free(attribute.attribute)

    cdef Py_ssize_t free_persistent(self, Attribute *attribute) noexcept nogil:
        cdef MetricSubsequence *ms
        if attribute.attribute != NULL:
            ms = <MetricSubsequence*> attribute.attribute
            self.metrics.free_persistent(ms.metric, &ms.subsequence)
            free(attribute.attribute)

    cdef Py_ssize_t init_persistent(
        self,
        TSArray X,
        Attribute *transient,
        Attribute *persistent
    ) noexcept nogil:
        cdef MetricSubsequenceView *msv = <MetricSubsequenceView*> transient.attribute
        cdef MetricSubsequence *ms = <MetricSubsequence*> malloc(
            sizeof(MetricSubsequence)
        )

        self.metrics.init_persistent(msv.metric, X, &msv.view, &ms.subsequence)
        ms.metric = msv.metric

        persistent.dim = transient.dim
        persistent.attribute = ms
        return 1

    cdef double transient_value(
        self, Attribute *attribute, TSArray X, Py_ssize_t sample
    ) noexcept nogil:
        cdef MetricSubsequenceView *msv = <MetricSubsequenceView*> attribute.attribute
        return self.metrics.transient_distance(
            msv.metric, &msv.view, X, sample, NULL
        )

    cdef double persistent_value(
        self, Attribute *attribute, TSArray X, Py_ssize_t sample
    ) noexcept nogil:
        cdef MetricSubsequence *ms = <MetricSubsequence*> attribute.attribute
        return self.metrics.persistent_distance(
            ms.metric, &ms.subsequence, X, sample, NULL
        )

    cdef Py_ssize_t transient_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t attribute_id,
    ) noexcept nogil:
        out[out_sample, attribute_id] = self.transient_value(attribute, X, sample)
        return 0

    cdef Py_ssize_t persistent_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t attribute_id,
    ) noexcept nogil:
        out[out_sample, attribute_id] = self.persistent_value(attribute, X, sample)
        return 0

    cdef object persistent_to_object(self, Attribute *attribute):
        cdef MetricSubsequence *ms = <MetricSubsequence*> attribute.attribute
        return attribute.dim, (
            ms.metric, self.metrics.to_array(ms.metric, &ms.subsequence)
        )

    cdef Py_ssize_t persistent_from_object(self, object object, Attribute *attribute):
        cdef MetricSubsequence *ms = <MetricSubsequence*> malloc(
            sizeof(MetricSubsequence)
        )
        dim, (metric, obj) = object
        self.metrics.from_array(metric, &ms.subsequence, obj)
        attribute.dim = dim
        attribute.attribute = ms
        return 0


cdef class RandomMultiMetricShapeletAttributeGenerator(
    MultiMetricShapeletAttributeGenerator
):
    cdef Py_ssize_t n_shapelets

    def __init__(
        self,
        Py_ssize_t n_shapelets,
        Py_ssize_t min_shapelet_size,
        Py_ssize_t max_shapelet_size,
        list metrics,
        const double[::1] weights=None,
    ):
        super().__init__(min_shapelet_size, max_shapelet_size, metrics, weights)
        self.n_shapelets = n_shapelets

    def __reduce__(self):
        return self.__class__, (
            self.n_shapelets,
            self.min_shapelet_size,
            self.max_shapelet_size,
            self.metrics.py_list,
            np.asarray(self.weights)
        )

    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil:
        return self.n_shapelets

    cdef Py_ssize_t next_attribute(
        self,
        Py_ssize_t attribute_id,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Attribute *transient,
        uint32_t *seed
    ) noexcept nogil:
        if attribute_id >= self.n_shapelets:
            return -1

        cdef Py_ssize_t shapelet_length
        cdef Py_ssize_t shapelet_start
        cdef Py_ssize_t shapelet_index
        cdef Py_ssize_t shapelet_dim
        cdef MetricSubsequenceView *msv = <MetricSubsequenceView*> malloc(
            sizeof(MetricSubsequenceView)
        )

        shapelet_length = rand_int(self.min_shapelet_size, self.max_shapelet_size, seed)
        shapelet_start = rand_int(0, X.shape[2] - shapelet_length, seed)
        shapelet_index = samples[rand_int(0, n_samples, seed)]
        if X.shape[1] > 1:
            shapelet_dim = rand_int(0, X.shape[1], seed)
        else:
            shapelet_dim = 0

        if self.weighted:
            msv.metric = vose_rand_int(&self.vr, seed)
        else:
            msv.metric = rand_int(0, self.metrics.size, seed)

        transient.dim = shapelet_dim
        self.metrics.init_transient(
            msv.metric,
            X,
            &msv.view,
            shapelet_index,
            shapelet_start,
            shapelet_length,
            shapelet_dim,
        )
        transient.attribute = msv
        return 1


cdef struct DilatedShapelet:
    Py_ssize_t dilation
    Py_ssize_t padding
    Py_ssize_t length
    double* data
    bint is_norm
    double threshold


cdef class DilatedShapeletAttributeGenerator(AttributeGenerator):
    cdef Py_ssize_t n_shapelets
    cdef Py_ssize_t *shapelet_length
    cdef Py_ssize_t n_shapelet_length
    cdef double norm_prob
    cdef double lower_bound
    cdef double upper_bound

    cdef Metric metric
    cdef double* x_buffer
    cdef double* k_buffer
    cdef double* dist_buffer
    cdef Py_ssize_t *arg_buffer

    cdef const Py_ssize_t[:] labels
    cdef const Py_ssize_t[:] samples
    cdef const Py_ssize_t[:] samples_per_label

    def __cinit__(
        self,
        Metric metric,
        Py_ssize_t n_shapelets,
        np.ndarray shapelet_length,
        double norm_prob,
        double lower_bound,
        double upper_bound,
        const Py_ssize_t[:] labels,   # must be the output of np.unique(return_indices=True)
        const Py_ssize_t[:] samples,  # must be np.argsort(labels)
        const Py_ssize_t[:] samples_per_label,  # must be the output of np.unique(return_count=True)
    ):
        self.metric = metric
        self.n_shapelets = n_shapelets
        self.norm_prob = norm_prob
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.n_shapelet_length = shapelet_length.shape[0]
        self.shapelet_length = <Py_ssize_t*> malloc(
            sizeof(Py_ssize_t) * self.n_shapelet_length
        )
        self.x_buffer = NULL
        self.k_buffer = NULL
        self.arg_buffer = NULL
        self.dist_buffer = NULL

        self.labels = labels
        self.samples = samples
        self.samples_per_label = samples_per_label

        cdef Py_ssize_t i
        for i in range(self.n_shapelet_length):
            self.shapelet_length[i] = shapelet_length[i]

    def __reduce__(self):
        return self.__class__, (
            self.metric,
            self.n_shapelets,
            to_ndarray_int(self.shapelet_length, self.n_shapelet_length),
            self.norm_prob,
            self.lower_bound,
            self.upper_bound,
            self.labels.base if self.labels is not None else None,
            self.samples.base if self.samples is not None else None,
            self.samples_per_label.base if self.samples_per_label is not None else None,
        )

    def __dealloc__(self):
        self._free()
        if self.shapelet_length != NULL:
            free(self.shapelet_length)
            self.shapelet_length = NULL


    cdef void _free(self) noexcept nogil:
        if self.x_buffer != NULL:
            free(self.x_buffer)
            self.x_buffer = NULL

        if self.k_buffer != NULL:
            free(self.k_buffer)
            self.k_buffer = NULL

        if self.dist_buffer != NULL:
            free(self.dist_buffer)
            self.dist_buffer = NULL

        if self.arg_buffer != NULL:
            free(self.arg_buffer)
            self.arg_buffer = NULL

    cdef int reset(self, TSArray X) noexcept nogil:
        self._free()
        self.x_buffer = <double*> malloc(sizeof(double) * X.shape[2] + 1)
        self.k_buffer = <double*> malloc(sizeof(double) * X.shape[2] + 1)
        self.dist_buffer = <double*> malloc(sizeof(double) * X.shape[2] + 1)
        self.arg_buffer = <Py_ssize_t*> malloc(sizeof(double) * X.shape[2] + 1)
        self.metric.reset(X, X)

        return 1

    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil:
        return self.n_shapelets

    cdef Py_ssize_t get_n_outputs(self, TSArray X) noexcept nogil:
        return self.get_n_attributess(X) * 3  # min, argmin, SO(S, X)

    cdef Py_ssize_t next_attribute(
        self,
        Py_ssize_t attribute_id,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Attribute *transient,
        uint32_t *seed
    ) noexcept nogil:
        cdef double mean, std
        cdef Py_ssize_t i, start, index, dim
        cdef DilatedShapelet *shapelet = <DilatedShapelet*> malloc(
            sizeof(DilatedShapelet)
        )

        shapelet.length = self.shapelet_length[
            rand_int(0, self.n_shapelet_length, seed)
        ]

        # TODO: ensure that the shapelet length is odd
        # so that the dilation/padding works as expected. Drop +1 above.
        # if shapelet.length > X.shape[2]:
        #     shapelet.length = X.shape[2]

        shapelet.data = <double*> malloc(sizeof(double) * shapelet.length)

        if X.shape[2] > 1:
            shapelet.dilation = <Py_ssize_t> floor(
                pow(
                    2,
                    rand_uniform(
                        0, max(0, log2((X.shape[2] - 1) / (shapelet.length - 1))), seed
                    ),
                )
            )
        else:
            shapelet.dilation = 1

        cdef Py_ssize_t dilated_length = (shapelet.length - 1) * shapelet.dilation + 1
        start = rand_int(0, max(0, X.shape[2] - dilated_length), seed)
        index = samples[rand_int(0, n_samples, seed)]
        if X.shape[1] > 1:
            dim = rand_int(0, X.shape[1], seed)
        else:
            dim = 0

        shapelet.padding = dilated_length // 2
        shapelet.is_norm = rand_uniform(0, 1, seed) < self.norm_prob

        # cdef const double* sample = &X[index, dim, start]
        cdef Py_ssize_t j = 0
        for i from start <= i < start + dilated_length by shapelet.dilation:
            if i < X.shape[2]:
                shapelet.data[j] = X[index, dim, i]
            else:
                shapelet.data[j] = 0
            j += 1

        if shapelet.is_norm:
            fast_mean_std(shapelet.data, shapelet.length, &mean, &std)
            if std == 0:
                std = 1

            for i in range(shapelet.length):
                shapelet.data[i] = (shapelet.data[i] - mean) / std

        if self.samples is not None:
            index = self._sample_other_same_label(index, seed)

        cdef Py_ssize_t n_distances = self._get_distance_profile(
            shapelet, &X[index, dim, 0], X.shape[2], False
        )

        for i in range(n_distances):
            self.arg_buffer[i] = i
        argsort(self.dist_buffer, self.arg_buffer, n_distances)

        cdef Py_ssize_t lower = <Py_ssize_t> floor(n_distances * self.lower_bound)
        cdef Py_ssize_t upper = <Py_ssize_t> floor(n_distances * self.upper_bound)
        cdef Py_ssize_t rand_threshold = self.arg_buffer[rand_int(lower, upper, seed)]
        shapelet.threshold = self.dist_buffer[rand_threshold]
        transient.dim = dim
        transient.attribute = shapelet
        return 0


    cdef Py_ssize_t free_transient(self, Attribute *attribute) noexcept nogil:
        return self.free_persistent(attribute)

    cdef Py_ssize_t free_persistent(self, Attribute *attribute) noexcept nogil:
        cdef DilatedShapelet *shapelet
        if attribute.attribute != NULL:
            shapelet = <DilatedShapelet*> attribute.attribute
            if shapelet.data != NULL:
                free(shapelet.data)
            free(attribute.attribute)
            attribute.attribute = NULL
        return 0

    # NOTE: We move ownership of `transient.attribute` to `persistent.attribute`.
    cdef Py_ssize_t init_persistent(
        self,
        TSArray X,
        Attribute *transient,
        Attribute *persistent
    ) noexcept nogil:
        persistent.dim = transient.dim
        persistent.attribute = transient.attribute
        transient.attribute = NULL  # No need to free it
        return 0

    cdef object persistent_to_object(self, Attribute *attribute):
        cdef Py_ssize_t j
        cdef DilatedShapelet *shapelet = <DilatedShapelet*> attribute.attribute

        data = np.empty(shapelet.length, dtype=float)
        for j in range(shapelet.length):
            data[j] = shapelet.data[j]

        return attribute.dim, (
            shapelet.dilation,
            shapelet.padding,
            shapelet.length,
            data,
            shapelet.is_norm,
            shapelet.threshold,
        )

    cdef Py_ssize_t persistent_from_object(self, object object, Attribute *attribute):
        dim, (dilation, padding, length, data, is_norm, threshold) = object

        cdef DilatedShapelet *shapelet = <DilatedShapelet*> malloc(sizeof(DilatedShapelet))
        shapelet.dilation = dilation
        shapelet.padding = padding
        shapelet.length = length
        shapelet.data = <double*> malloc(sizeof(double) * length)
        shapelet.threshold = threshold
        shapelet.is_norm = is_norm

        cdef Py_ssize_t i
        for i in range(length):
            shapelet.data[i] = data[i]

        attribute.attribute = shapelet
        attribute.dim = dim
        return 0

    cdef Py_ssize_t transient_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t attribute_id,
    ) noexcept nogil:
        cdef double minval = INFINITY
        cdef double minarg = -1
        cdef double so = 0
        cdef Py_ssize_t i
        cdef DilatedShapelet* shapelet = <DilatedShapelet*> attribute.attribute

        cdef Py_ssize_t n_distances = self._get_distance_profile(
            shapelet, &X[sample, attribute.dim, 0], X.shape[2], False
        )

        # If we got no distances, we need to compute the full
        # distance profile to find the min.
        if n_distances == 0:
            n_distances = self._get_distance_profile(
                shapelet, &X[sample, attribute.dim, 0], X.shape[2], False
            )

        for i in range(n_distances):
            if self.dist_buffer[i] < minval:
                minval = self.dist_buffer[i]
                minarg = i
            if self.dist_buffer[i] < shapelet.threshold:
                so += 1

        cdef Py_ssize_t attribute_offset = attribute_id * 3
        out[out_sample, attribute_offset] = minval
        out[out_sample, attribute_offset + 1] = minarg
        out[out_sample, attribute_offset + 2] = so
        return 0

    cdef Py_ssize_t persistent_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t out_attribute,
    ) noexcept nogil:
        return self.transient_fill(
            attribute, X, sample, out, out_sample, out_attribute
        )

    cdef double transient_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil:
        cdef double minval = INFINITY
        cdef Py_ssize_t so = 0
        cdef DilatedShapelet* shapelet = <DilatedShapelet*> attribute.attribute
        cdef Py_ssize_t n_distances = self._get_distance_profile(
            shapelet, &X[sample, attribute.dim, 0], X.shape[2], True
        )
        if n_distances == 0:
            n_distances = self._get_distance_profile(
                shapelet, &X[sample, attribute.dim, 0], X.shape[2], False
            )

        for i in range(n_distances):
            if self.dist_buffer[i] < minval:
                minval = self.dist_buffer[i]
            if self.dist_buffer[i] < shapelet.threshold:
                so += 1

        return minval / so

    cdef double persistent_value(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample
    ) noexcept nogil:
        return self.transient_value(attribute, X, sample)

    # Guarded by if self.samples is not None
    cdef Py_ssize_t _sample_other_same_label(
        self, Py_ssize_t sample, uint32_t *seed
    ) noexcept nogil:
        cdef Py_ssize_t label = self.labels[sample]
        cdef Py_ssize_t label_start = 0
        cdef Py_ssize_t label_end = 0
        cdef Py_ssize_t i
        for i in range(label):
            label_start += self.samples_per_label[i]
            label_end += self.samples_per_label[i]

        label_end += self.samples_per_label[label]
        return self.samples[rand_int(label_start, label_end, seed)]

    cdef Py_ssize_t _get_distance_profile(
        self,
        DilatedShapelet* shapelet,
        const double* x,
        Py_ssize_t x_len,
        bint early_abandon,
    ) noexcept nogil:
        if shapelet.is_norm:
            return scaled_dilated_distance_profile(
                1,
                shapelet.dilation,
                shapelet.padding,
                shapelet.data,
                shapelet.length,
                x,
                x_len,
                self.metric,
                self.x_buffer,
                self.k_buffer,
                shapelet.threshold if early_abandon else INFINITY,
                self.dist_buffer,
            )
        else:
            return dilated_distance_profile(
                1,
                shapelet.dilation,
                shapelet.padding,
                shapelet.data,
                shapelet.length,
                x,
                x_len,
                self.metric,
                self.x_buffer,
                self.k_buffer,
                shapelet.threshold if early_abandon else INFINITY,
                self.dist_buffer,
            )


cdef inline Py_ssize_t _max_exponent(
    Py_ssize_t n_timestep, Py_ssize_t kernel_size
) noexcept nogil:
    cdef Py_ssize_t max_exponent = <Py_ssize_t> floor(
        log2((n_timestep - 1) / <double> (kernel_size - 1))
    )
    if max_exponent < 0:
        max_exponent = 0
    max_exponent += 1

    return max_exponent


cdef struct Cdist:
    bint is_norm
    Py_ssize_t max_exponent

    # size: max_exponent * n_shapelets
    double *thresholds

    # size: shapelet_size * n_shapelets
    double *data


cdef class CastorSummarizer:
    cdef Py_ssize_t n_timestep
    cdef Py_ssize_t n_shapelets
    cdef bint soft_min, soft_max, soft_threshold
    cdef double *min_values
    cdef double *max_values
    cdef double *min_so_values

    def __cinit__(self, *args, **kwargs):
        self.min_values = NULL
        self.max_values = NULL
        self.min_so_values = NULL

    def __init__(self, soft_min=True, soft_max=False, soft_threshold=True):
        self.soft_min = soft_min
        self.soft_max = soft_max
        self.soft_threshold = soft_threshold

    def __reduce__(self):
        return self.__class__, (self.soft_min, self.soft_max, self.soft_threshold)

    def __dealloc__(self):
        self._free()

    cdef Py_ssize_t get_n_features(self) noexcept nogil:
        return 3

    cdef void _free(self) noexcept nogil:
        if self.min_values != NULL:
            free(self.min_values)
            self.min_values = NULL
        if self.max_values != NULL:
            free(self.max_values)
            self.max_values = NULL
        if self.min_so_values != NULL:
            free(self.min_so_values)
            self.min_so_values = NULL

    cdef int reset(self, Py_ssize_t n_timestep, Py_ssize_t n_shapelets) noexcept nogil:
        self._free()
        self.n_timestep = n_timestep
        self.n_shapelets = n_shapelets
        self.min_values = <double*> malloc(sizeof(double) * self.n_shapelets)
        self.max_values = <double*> malloc(sizeof(double) * self.n_shapelets)
        self.min_so_values = <double*> malloc(sizeof(double) * self.n_shapelets)

    cdef void fill(
        self,
        Py_ssize_t attribute_offset,
        Py_ssize_t exponent,
        Cdist *cdist,
        double *dist_buffer,
        double[:] out,
    ) noexcept nogil:
        cdef Py_ssize_t i, min_index, max_index
        cdef double min_value, max_value

        memset(self.min_values, 0, sizeof(double) * self.n_shapelets)
        memset(self.max_values, 0, sizeof(double) * self.n_shapelets)
        memset(self.min_so_values, 0, sizeof(double) * self.n_shapelets)

        for i in range(self.n_timestep):
            self._compute_attributes(
                i,
                self.n_timestep,
                exponent,
                cdist.max_exponent,
                dist_buffer,
                cdist.thresholds,
                &min_index,
                &min_value,
                &max_index,
                &max_value,
                self.min_so_values,
            )

            self.max_values[max_index] += max_value if self.soft_max else 1
            self.min_values[min_index] += min_value if self.soft_min else 1
            if (
                not self.soft_threshold and
                self.min_values[min_index] < cdist.thresholds[min_index * cdist.max_exponent + exponent]
            ):
                self.min_so_values[min_index] += 1

        cdef Py_ssize_t shapelet_attribute_offset = attribute_offset + exponent * self.n_shapelets * 3
        for i in range(self.n_shapelets):
            out[shapelet_attribute_offset + i * 3 + 0] = self.min_values[i]
            out[shapelet_attribute_offset + i * 3 + 1] = self.max_values[i]
            out[shapelet_attribute_offset + i * 3 + 2] = self.min_so_values[i]

    cdef void _compute_attributes(
        self,
        Py_ssize_t offset,
        Py_ssize_t stride,
        Py_ssize_t exponent,
        Py_ssize_t max_exponent,
        double *values,
        double *thresholds,
        Py_ssize_t *min_index,
        double *min_value,
        Py_ssize_t *max_index,
        double *max_value,
        double *min_so_values,
    ) noexcept nogil:
        cdef Py_ssize_t i
        cdef double value, threshold
        min_value[0] = INFINITY
        max_value[0] = -INFINITY
        for i in range(self.n_shapelets):
            value = values[offset + i * stride]
            threshold = thresholds[i * max_exponent + exponent]

            if value < min_value[0]:
                min_value[0] = value
                min_index[0] = i
            if value < threshold and self.soft_threshold:
                min_so_values[i] += 1
            if value > max_value[0]:
                max_value[0] = value
                max_index[0] = i

cdef class CastorAttributeGenerator(AttributeGenerator):
    cdef Py_ssize_t n_shapelets
    cdef Py_ssize_t shapelet_size
    cdef Py_ssize_t n_groups
    cdef double normalize_prob
    cdef double lower
    cdef double upper

    cdef CastorSummarizer summarizer
    cdef Metric metric

    # Temporary buffer to store dilated distance
    cdef double *dist_buffer
    cdef Py_ssize_t *arg_buffer

    cdef double *x_buffer
    cdef double *s_buffer

    # Temporary buffer to store kernel values
    cdef double *max_values
    cdef double *min_values
    cdef double *min_so_values
    cdef double *max_so_values

    # random values and incremental indices to draw random samples from
    cdef double *rnd_value

    cdef const Py_ssize_t[:] labels
    cdef const Py_ssize_t[:] samples
    cdef const Py_ssize_t[:] samples_per_label

    def __cinit__(
        self,
        Py_ssize_t n_groups,
        Py_ssize_t n_shapelets,
        Py_ssize_t shapelet_size,
        double normalize_prob,
        double lower,
        double upper,
        Metric metric,
        CastorSummarizer summarizer,
        const Py_ssize_t[:] labels,   # must be the output of np.unique(return_indices=True)
        const Py_ssize_t[:] samples,  # must be np.argsort(labels)
        const Py_ssize_t[:] samples_per_label,  # must be the output of np.unique(return_count=True)
    ):
        self.n_groups = n_groups
        self.n_shapelets = n_shapelets
        self.shapelet_size = shapelet_size
        self.normalize_prob = normalize_prob
        self.lower = lower
        self.upper = upper
        self.metric = metric
        self.summarizer = summarizer

        self.rnd_value = NULL
        self.dist_buffer = NULL
        self.arg_buffer = NULL
        self.x_buffer = NULL
        self.s_buffer = NULL
        self.max_values = NULL
        self.min_values = NULL
        self.min_so_values = NULL
        self.max_so_values = NULL
        self.labels = labels
        self.samples = samples
        self.samples_per_label = samples_per_label

    def __reduce__(self):
        return self.__class__, (
            self.n_groups,
            self.n_shapelets,
            self.shapelet_size,
            self.normalize_prob,
            self.lower,
            self.upper,
            self.metric,
            self.summarizer,
            self.labels.base if self.labels is not None else None,
            self.samples.base if self.samples is not None else None,
            self.samples_per_label.base if self.samples_per_label is not None else None,
        )

    def __dealloc__(self):
        self._free()

    cdef void _free(self) noexcept nogil:
        if self.arg_buffer != NULL:
            free(self.arg_buffer)
            self.arg_buffer = NULL
        if self.dist_buffer != NULL:
            free(self.dist_buffer)
            self.dist_buffer = NULL
        if self.x_buffer != NULL:
            free(self.x_buffer)
            self.x_buffer = NULL
        if self.s_buffer != NULL:
            free(self.s_buffer)
            self.s_buffer = NULL
        if self.min_values != NULL:
            free(self.min_values)
            self.min_values = NULL
        if self.max_values != NULL:
            free(self.max_values)
            self.max_values = NULL
        if self.min_so_values != NULL:
            free(self.min_so_values)
            self.min_so_values = NULL
        if self.max_so_values != NULL:
            free(self.max_so_values)
            self.max_so_values = NULL
        if self.rnd_value != NULL:
            free(self.rnd_value)
            self.rnd_value = NULL

    cdef int reset(self, TSArray X) noexcept nogil:
        self._free()
        self.dist_buffer = <double*> malloc(
            sizeof(double) * X.shape[2] * self.n_shapelets
        )
        self.arg_buffer = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * X.shape[2])
        self.x_buffer = <double*> malloc(sizeof(double) * self.shapelet_size)
        self.s_buffer = <double*> malloc(sizeof(double) * self.shapelet_size)
        self.min_values = <double*> malloc(sizeof(double) * self.n_shapelets)
        self.max_values = <double*> malloc(sizeof(double) * self.n_shapelets)
        self.min_so_values = <double*> malloc(sizeof(double) * self.n_shapelets)
        self.max_so_values = <double*> malloc(sizeof(double) * self.n_shapelets)
        self.rnd_value = <double*> malloc(sizeof(double) * X.shape[0])

        self.metric.reset(X, X)
        self.summarizer.reset(X.shape[2], self.n_shapelets)
        return 0

    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil:
        return self.n_groups

    # n_shapelets * max_exponent * n_shapelets * 3 features per time series.
    cdef Py_ssize_t get_n_outputs(self, TSArray X) noexcept nogil:
        cdef Py_ssize_t max_exponent = _max_exponent(X.shape[2], self.shapelet_size)
        return (
            self.get_n_attributess(X) * max_exponent * self.n_shapelets * self.summarizer.get_n_features()
        )

    cdef Py_ssize_t _get_distance_profile(
        self,
        Py_ssize_t dilation,
        Py_ssize_t padding,
        const double *s,
        Py_ssize_t s_len,
        const double *x,
        Py_ssize_t x_len,
        bint is_norm,
        double *dist_buffer,
    ) noexcept nogil:
        if is_norm:
            return scaled_dilated_distance_profile(
                1,
                dilation,
                padding,
                s,
                s_len,
                x,
                x_len,
                self.metric,
                self.x_buffer,
                self.s_buffer,
                INFINITY,
                dist_buffer,
            )
        else:
            return dilated_distance_profile(
                1,
                dilation,
                padding,
                s,
                s_len,
                x,
                x_len,
                self.metric,
                self.x_buffer,
                self.s_buffer,
                INFINITY,
                dist_buffer,
            )

    cdef Py_ssize_t next_attribute(
        self,
        Py_ssize_t attribute_id,
        TSArray X,
        Py_ssize_t *samples,
        Py_ssize_t n_samples,
        Attribute *transient,
        uint32_t *seed
    ) noexcept nogil:
        cdef Cdist *cdist = <Cdist*> malloc(sizeof(Cdist))
        cdef Py_ssize_t i, j, k, m, n
        cdef Py_ssize_t dim, start, dilation, dilated_size, padding, n_distances
        cdef Py_ssize_t lower, upper
        cdef double mean, std
        cdef Py_ssize_t max_exponent = _max_exponent(X.shape[2], self.shapelet_size)

        for i in range(n_samples):
            self.rnd_value[i] = rand_uniform(0, 1, seed)

        argsort(self.rnd_value, samples, n_samples)

        # shapelet[0, 0], ..., shapelet[n_shapelets, max_exponent]
        cdef double *data = <double*> malloc(
            sizeof(double) * self.n_shapelets * max_exponent * self.shapelet_size
        )
        cdef double *thresholds = <double*> malloc(
            sizeof(double) * max_exponent * self.n_shapelets
        )
        cdef double *shapelet
        cdef double *x

        if X.shape[1] > 1:
            dim = rand_int(0, X.shape[1], seed)
        else:
            dim = 0

        cdist.is_norm = rand_uniform(0, 1, seed) < self.normalize_prob
        cdist.max_exponent = max_exponent

        # Draw samples per shapelet/dilation pair not per shapelet.
        for i in range(self.n_shapelets):
            for k in range(max_exponent):
                j = samples[(i + k) % n_samples]
                shapelet = (
                    data
                    + i * self.shapelet_size * max_exponent
                    + k * self.shapelet_size
                )

                dilation = <Py_ssize_t> pow(2, k)
                dilated_size = (self.shapelet_size - 1) * dilation + 1
                start = rand_int(0, max(0, X.shape[2] - dilated_size), seed)
                padding = dilated_size // 2

                n = 0
                for m from start <= m < start + dilated_size by dilation:
                    if m < X.shape[2]:
                        shapelet[n] = X[j, dim, m]
                    else:
                        shapelet[n] = 0.0
                    n += 1

                if cdist.is_norm:
                    fast_mean_std(shapelet, self.shapelet_size, &mean, &std)
                    if std == 0.0:
                        std = 1.0

                    for m in range(self.shapelet_size):
                        shapelet[m] = (shapelet[m] - mean) / std

                if self.samples is not None:
                    x = &X[self._sample_other_same_label(j, seed), dim, 0]
                else:
                    x = &X[j, dim, 0]

                n_distances = self._get_distance_profile(
                    dilation,
                    padding,
                    shapelet,
                    self.shapelet_size,
                    x,
                    X.shape[2],
                    cdist.is_norm,
                    self.dist_buffer,
                )

                for m in range(n_distances):
                    self.arg_buffer[m] = m
                argsort(self.dist_buffer, self.arg_buffer, n_distances)

                lower = <Py_ssize_t> floor(n_distances * self.lower)
                upper = <Py_ssize_t> floor(n_distances * self.upper)

                thresholds[i * max_exponent + k] = self.dist_buffer[
                    rand_int(lower, upper, seed)
                ]

        cdist.thresholds = thresholds
        cdist.data = data

        transient.dim = dim
        transient.attribute = cdist
        return 0

    # NOTE: We move ownership of `transient.attribute` to `persistent.attribute`.
    cdef Py_ssize_t init_persistent(
        self,
        TSArray X,
        Attribute *transient,
        Attribute *persistent
    ) noexcept nogil:
        persistent.dim = transient.dim
        persistent.attribute = transient.attribute
        transient.attribute = NULL  # No need to free it
        return 0

    cdef Py_ssize_t free_persistent(self, Attribute *attribute) noexcept nogil:
        cdef Cdist *cdist
        if attribute.attribute != NULL:
            cdist = <Cdist*> attribute.attribute
            if cdist.data != NULL:
                free(cdist.data)
            if cdist.thresholds != NULL:
                free(cdist.thresholds)

            free(attribute.attribute)
            attribute.attribute = NULL
        return 0

    cdef Py_ssize_t free_transient(self, Attribute *attribute) noexcept nogil:
        return self.free_persistent(attribute)

    cdef object persistent_to_object(self, Attribute *attribute):
        cdef Py_ssize_t i
        cdef Cdist *cdist = <Cdist*> attribute.attribute

        data = np.empty(
            self.shapelet_size * cdist.max_exponent * self.n_shapelets, dtype=float
        )
        thresholds = np.empty(cdist.max_exponent * self.n_shapelets, dtype=float)

        for i in range(data.shape[0]):
            data[i] = cdist.data[i]

        for i in range(thresholds.shape[0]):
            thresholds[i] = cdist.thresholds[i]

        return attribute.dim, (
            cdist.is_norm,
            cdist.max_exponent,
            data,
            thresholds,
        )

    cdef Py_ssize_t persistent_from_object(self, object obj, Attribute *attribute):
        dim, (is_norm, max_exponent, data, thresholds) = obj
        cdef Cdist *cdist = <Cdist*> malloc(sizeof(Cdist))

        cdist.is_norm = is_norm
        cdist.max_exponent = max_exponent
        cdist.data = <double*> malloc(sizeof(double) * data.shape[0])
        cdist.thresholds = <double*> malloc(sizeof(double) * thresholds.shape[0])
        cdef Py_ssize_t i

        for i in range(data.shape[0]):
            cdist.data[i] = data[i]

        for i in range(thresholds.shape[0]):
            cdist.thresholds[i] = thresholds[i]

        attribute.attribute = cdist
        attribute.dim = dim
        return 0

    cdef Py_ssize_t transient_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t attribute_id,
    ) noexcept nogil:
        cdef Py_ssize_t i, exponent, padding, dilation
        cdef double min_value, max_value
        cdef Py_ssize_t min_index, max_index

        cdef Cdist *cdist = <Cdist*> attribute.attribute
        cdef double *shapelet

        # Place the pointer inside correct attribute group, as given by attribute_id
        cdef Py_ssize_t shapelet_attribute_offset
        cdef Py_ssize_t attribute_offset = (
            attribute_id * self.n_shapelets * cdist.max_exponent * 3
        )

        for exponent in range(cdist.max_exponent):
            dilation = <Py_ssize_t> pow(2, exponent)
            padding = <Py_ssize_t> ((self.shapelet_size - 1) * dilation) // 2

            for i in range(self.n_shapelets):
                shapelet = (
                    cdist.data
                    + i * self.shapelet_size * cdist.max_exponent
                    + exponent * self.shapelet_size
                )
                self._get_distance_profile(
                    dilation,
                    padding,
                    shapelet,
                    self.shapelet_size,
                    &X[sample, attribute.dim, 0],
                    X.shape[2],
                    cdist.is_norm,
                    self.dist_buffer + i * X.shape[2],
                )

            self.summarizer.fill(attribute_offset, exponent, cdist, self.dist_buffer, out[out_sample])

    cdef Py_ssize_t persistent_fill(
        self,
        Attribute *attribute,
        TSArray X,
        Py_ssize_t sample,
        double[:, :] out,
        Py_ssize_t out_sample,
        Py_ssize_t out_attribute,
    ) noexcept nogil:
        return self.transient_fill(
            attribute, X, sample, out, out_sample, out_attribute
        )

    # Guarded by if self.samples is not None
    cdef Py_ssize_t _sample_other_same_label(
        self, Py_ssize_t sample, uint32_t *seed
    ) noexcept nogil:
        cdef Py_ssize_t label = self.labels[sample]
        cdef Py_ssize_t label_start = 0
        cdef Py_ssize_t label_end = 0
        cdef Py_ssize_t i
        for i in range(label):
            label_start += self.samples_per_label[i]
            label_end += self.samples_per_label[i]

        label_end += self.samples_per_label[label]
        return self.samples[rand_int(label_start, label_end, seed)]

    cdef void _compute_attributes(
        self,
        Py_ssize_t offset,
        Py_ssize_t stride,
        Py_ssize_t exponent,
        Py_ssize_t max_exponent,
        double *values,
        double *thresholds,
        Py_ssize_t *min_index,
        double *min_value,
        Py_ssize_t *max_index,
        double *max_value,
        double *min_so_values,
    ) noexcept nogil:
        cdef Py_ssize_t i
        cdef double value, threshold
        min_value[0] = INFINITY
        max_value[0] = -INFINITY
        for i in range(self.n_shapelets):
            value = values[offset + i * stride]
            threshold = thresholds[i * max_exponent + exponent]

            if value < min_value[0]:
                min_value[0] = value
                min_index[0] = i
            if value < threshold:
                min_so_values[i] += 1
            if value > max_value[0]:
                max_value[0] = value
                max_index[0] = i



import numpy as np
import pytest
from wildboar.datasets import load_gun_point
from wildboar.ensemble import ElasticEnsembleClassifier


def test_elastic_ensemble_classifier_fit():
    X, y = load_gun_point()
    X_train = X[:30]
    y_train = y[:30]
    clf = ElasticEnsembleClassifier(
        n_neighbors=1,
        metric={
            "dtw": {"min_r": 0.1, "max_r": 0.3, "num_r": 3},
            "ddtw": {"min_r": 0.1, "max_r": 0.3, "num_r": 3},
        },
    )
    clf.fit(X_train, y_train)
    proba = clf.predict_proba(X[30:40])
    np.testing.assert_almost_equal([score[1] for score in clf.scores_], [0.9, 1.0])
    np.testing.assert_equal(
        proba.argmax(axis=1),
        [0, 1, 1, 0, 1, 0, 1, 1, 1, 1],
    )



# Authors: Isak Samsten
# License: BSD 3 clause
import numpy as np
import pytest

from wildboar.ensemble import ExtraShapeletTreesClassifier, ExtraShapeletTreesRegressor
from wildboar.utils.estimator_checks import check_estimator


def test_shap_for():
    clf = ExtraShapeletTreesClassifier(n_estimators=1)
    clf.fit(np.zeros((10, 10)), np.ones(10))


def test_check_estimator():
    check_estimator(
        ExtraShapeletTreesClassifier(), ignore=["check_sample_weights_invariance"]
    )
    check_estimator(
        ExtraShapeletTreesRegressor(), ignore=["check_sample_weights_invariance"]
    )


@pytest.mark.parametrize(
    "estimator, expected_estimator_params",
    [
        (
            ExtraShapeletTreesClassifier(),
            (
                "max_depth",
                "min_samples_leaf",
                "min_impurity_decrease",
                "min_samples_split",
                "min_shapelet_size",
                "max_shapelet_size",
                "metric",
                "metric_params",
                "criterion",
            ),
        ),
        (
            ExtraShapeletTreesRegressor(),
            (
                "max_depth",
                "min_impurity_decrease",
                "min_samples_leaf",
                "min_samples_split",
                "min_shapelet_size",
                "max_shapelet_size",
                "metric",
                "metric_params",
                "criterion",
            ),
        ),
    ],
)
def test_estimator_params(estimator, expected_estimator_params):
    assert set(estimator.estimator_params) == set(expected_estimator_params)



# Authors: Isak Samsten
# License: BSD 3 clause

import base64
import hashlib
import json
import os
import pickle
import re
import sys
import warnings
import zipfile
from abc import ABCMeta, abstractmethod
from urllib.error import URLError
from urllib.request import urlopen, urlretrieve

import numpy as np
from sklearn.utils.fixes import parse_version

from .. import EOS
from .. import __version__ as wildboar_version

DEFAULT_TAG = "default"


def _replace_placeholders(url, **kwargs):
    """
    Replace placeholder values of the format {key} with kwargs[key].

    Parameters
    ----------
    url : str
        The input string
    **kwargs : dict
        The key and values to replace

    Returns
    -------
    str : url with placeholder values replaced
    """
    for arg, value in kwargs.items():
        url = url.replace("{%s}" % arg, value)
    return url


def _check_integrity(bundle_file, hash_file, throws=True):
    """
    Check the integrity of the downloaded or cached file.

    Parameters
    ----------
    bundle_file : str, bytes or PathLike
        Path to the bundle file
    hash_file : str, bytes or PathLike
        Path to the hash file
    throws : bool
        Throw an exception on hash missmatch

    Returns
    -------
    bool
        true if the hash of bundle file matches the contents of the hash file
    """
    with open(hash_file, "r") as f:
        hash = f.readline().strip()

    if hash is not None:
        actual_hash = _sha1(bundle_file)
        if hash != actual_hash:
            if throws:
                raise ValueError(
                    "Integrity check failed, expected '%s', got '%s'."
                    % (hash, actual_hash)
                )
            else:
                return False
    return True


def _sha1_is_sane(hash_file):
    """
    Check the sanity of a hash file.

    Parameters
    ----------
    hash_file : str, bytes or PathLike
        The path to the hash file

    Returns
    -------
    bool
        Returns true if hash is 40 characters long; otherwise false.
    """
    with open(hash_file, "r") as f:
        return len(f.readline().strip()) == 40


def _load_archive(
    bundle_name,
    download_url,
    cache_dir,
    *,
    create_cache_dir=True,
    progress=True,
    force=False,
):
    """
    Load or download a bundle.

    Parameters
    ----------
    bundle_name : str
        The name of the cached file.
    download_url : str
        The download url to the bundle and hash file.
    cache_dir : str
        The cache directory.
    create_cache_dir : bool
        Create the cache directory if missing.
    progress : bool
        Show progress information
    force : bool
        Remove any cached files and force re-download.

    Returns
    -------
    zipfile.ZipFile
        A zip-archive with datasets.
    """
    if not os.path.exists(cache_dir):
        if create_cache_dir:
            os.makedirs(os.path.abspath(cache_dir), exist_ok=True)
        else:
            raise ValueError(
                "The output directory %s does not exist. "
                "Set create_cache_dir=True to create it." % cache_dir
            )

    cached_hash = os.path.join(cache_dir, "%s.sha1" % bundle_name)
    cached_bundle = os.path.join(cache_dir, "%s.zip" % bundle_name)

    if force:
        if os.path.exists(cached_hash):
            os.remove(cached_hash)
        if os.path.exists(cached_bundle):
            os.remove(cached_bundle)

    if not os.path.exists(cached_hash) or not _sha1_is_sane(cached_hash):
        _download_hash_file(cached_hash, "%s.sha1" % download_url, bundle_name)

    if os.path.exists(cached_bundle) and _check_integrity(
        cached_bundle, cached_hash, throws=False
    ):
        try:
            return zipfile.ZipFile(open(cached_bundle, "rb"))
        except zipfile.BadZipFile:
            os.remove(cached_bundle)

    _download_bundle_file(cached_bundle, "%s.zip" % download_url, bundle_name, progress)
    _check_integrity(cached_bundle, cached_hash)
    return zipfile.ZipFile(open(cached_bundle, "rb"))


def _download_hash_file(cached_hash, hash_url, filename):
    """
    Download the hash file.

    Parameters
    ----------
    cached_hash : str, bytes or PathLike
        The path to the cached hash.
    hash_url : str
        The download url.
    filename : str
        The filename of the bundle.
    """
    with open(cached_hash, "wb") as f:
        try:
            with urlopen(hash_url, timeout=5) as response:
                f.write(response.read())
        except URLError as e:
            os.remove(cached_hash)
            raise ValueError(
                "Downloading %s.sha1 failed. Try another version or tag." % filename
            ) from e


_SIZE_NAMES = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")


def _convert_size(byte_size, i=None):
    import math

    if byte_size == 0:
        return 0, 0

    if i is None:
        i = int(math.floor(math.log(byte_size, 1024)))
    p = math.pow(1024, i)
    s = round(byte_size / p, 2)
    return s, i


def _print_progress(
    current_iter,
    max_iter,
    *,
    prefix="",
    suffix="",
    progress=None,
    length=50,
    fill="█",
    unfill="-",
    end="\r",
    file=sys.stderr,
):
    def size_progress(current, total):
        total_size, total_type = _convert_size(total)
        current_size, _ = _convert_size(current, total_type)
        return f" {current_size}/{total_size} {_SIZE_NAMES[total_type]}"

    if progress is None:
        progress = size_progress

    progress = progress(float(current_iter), float(max_iter))
    fill_size = int(length * current_iter // max_iter)
    unfill_size = int(length - fill_size)
    bar = fill * fill_size + unfill * unfill_size
    print(f"\r{prefix}|{bar}|{progress}{suffix}", file=file, end=end)
    if current_iter >= max_iter:
        print(file=file)


def _download_bundle_file(cached_bundle, bundle_url, filename, progress):
    """
    Download the bundle.

    Parameters
    ----------
    cached_bundle : str, bytes or PathLike
        The path to the cached bundle
    bundle_url : str
        The download url
    filename : str
        The filename of the bundle
    progress : bool
        Show progress bar
    """

    if progress:

        def reporthook(blocknum, bs, size):
            _print_progress(blocknum * bs, size)
    else:
        reporthook = None

    try:
        urlretrieve(bundle_url, cached_bundle, reporthook=reporthook)
    except URLError as e:
        os.remove(cached_bundle)
        raise ValueError(
            "Downloading %s.zip failed. Try another version or tag." % filename
        ) from e


class Repository(metaclass=ABCMeta):
    """A repository is a collection of bundles."""

    def __init__(self):
        self._active = False

    def __eq__(self, o):
        if isinstance(o, Repository):
            return self.identifier == o.identifier
        return False

    def __hash__(self) -> int:
        return hash(self.identifier)

    @property
    @abstractmethod
    def identifier(self):
        pass

    @property
    @abstractmethod
    def name(self):
        """
        Name of the repository.

        Returns
        -------
        str
            The name of the repository.
        """
        pass

    @property
    @abstractmethod
    def version(self):
        """
        The repository version.

        Returns
        -------
        str
            The version of the repository.
        """
        pass

    @property
    @abstractmethod
    def download_url(self):
        """
        The url template for downloading bundles.

        Returns
        -------
        str
            The download url.
        """
        pass

    @property
    @abstractmethod
    def wildboar_requires(self):
        """
        The minimum required wildboar version.

        Returns
        -------
        str
            The min version.
        """
        pass

    @abstractmethod
    def get_bundles(self):
        """
        Get all bundles.

        Returns
        -------
        dict
            A dictionary of key and bundle.
        """
        pass

    @property
    def active(self):
        return self._active

    def get_bundle(self, key):
        """
        Get a bundle with the specified key.

        Parameters
        ----------
        key : str
            Key of the bundle.

        Returns
        -------
        Bundle, optional
            A bundle or None.
        """
        bundle = self.get_bundles().get(key)
        if bundle is None:
            raise ValueError("The bundle %s does not exist." % key)
        return bundle

    def load_dataset(
        self,
        bundle,
        dataset,
        *,
        cache_dir,
        tag=None,
        create_cache_dir=True,
        progress=True,
        force=False,
    ):
        bundle = self.get_bundle(bundle)
        version = bundle.version
        tag = tag or bundle.tag

        cache_dir = os.path.join(cache_dir, self.identifier)
        download_url = _replace_placeholders(
            self.download_url, bundle=bundle.key, version=version, tag=tag
        )

        with _load_archive(
            bundle_name=bundle.get_filename(version, tag),
            download_url=download_url,
            cache_dir=cache_dir,
            create_cache_dir=create_cache_dir,
            progress=progress,
            force=force,
        ) as archive:
            return bundle.load(dataset, archive)

    def list_datasets(
        self,
        bundle,
        *,
        cache_dir,
        collection=None,
        tag=None,
        create_cache_dir=True,
        progress=True,
        force=False,
    ):
        bundle = self.get_bundle(bundle)
        version = bundle.version
        tag = tag or bundle.tag

        cache_dir = os.path.join(cache_dir, self.identifier)
        download_url = _replace_placeholders(
            self.download_url, bundle=bundle.key, version=version, tag=tag
        )
        with _load_archive(
            bundle_name=bundle.get_filename(version, tag),
            download_url=download_url,
            cache_dir=cache_dir,
            create_cache_dir=create_cache_dir,
            progress=progress,
            force=force,
        ) as archive:
            return bundle.list(archive, collection)

    def clear_cache(self, cache_dir, keep_last_version=True):
        cache_dir = os.path.join(cache_dir, self.identifier)
        if not os.path.exists(cache_dir) or not os.path.isdir(cache_dir):
            return

        keep = []
        if keep_last_version:
            keep = [
                bundle.get_filename(tag=bundle.tag)
                for _, bundle in self.get_bundles().items()
            ]

        for filename in os.listdir(cache_dir):
            basename, ext = os.path.splitext(filename)
            full_path = os.path.join(cache_dir, filename)
            if (
                os.path.isfile(full_path)
                and ext in [".zip", ".sha1"]
                and basename not in keep
            ):
                os.remove(full_path)

    def refresh(self, timeout=None):
        """Refresh the repository."""
        try:
            self._refresh(timeout)
            self._active = True
        except URLError:
            self._active = False

    @abstractmethod
    def _refresh(self, timeout):
        pass


def _validate_url(url):
    if "{bundle}" in url and "{version}" in url and "{tag}" in url:
        return url
    else:
        raise ValueError(
            "The repository url is invalid, got %s "
            "({bundle}, {version} and {tag} are required)" % url
        )


def _validate_repository_name(str):
    if re.match("[a-zA-Z]+", str):
        return str
    else:
        raise ValueError(
            "The repository name %s is not valid, expected only characters." % str
        )


def _validate_bundle_key(str):
    if re.match(r"[a-zA-Z0-9\-]+", str):
        return str
    else:
        raise ValueError(
            "The bundle key %s is not valid, expected only characters, numbers or -."
            % str
        )


def _validate_version(str, *, max_version=None):
    if re.match(r"(^(?:\d+\.)?(?:\d+\.)?(?:\*|\d+)$)", str):
        if max_version and parse_version(str) > parse_version(max_version):
            raise ValueError(
                "The dataset is not supported, required version >=%s, got %s"
                % (max_version, str)
            )
        return str
    else:
        raise ValueError("The version %s is not valid." % str)


def _validate_collections(collections):
    if not isinstance(collections, dict):
        raise TypeError(
            "'collections' must be dict, not %r" % type(collections).__qualname__
        )
    else:
        for key, values in collections.items():
            if not isinstance(values, list):
                raise TypeError(
                    "'collections.value' must be list, not %r"
                    % type(values).__qualname__
                )
            if not isinstance(key, str):
                raise TypeError(
                    "'collections.key' must be str, not" % type(key).__qualname__
                )
    return collections


class JSONRepository(Repository):
    supported_version = "1.1"

    def __init__(self, url):
        super().__init__()
        self.repo_url = url

    @property
    def identifier(self):
        return base64.b64encode(self.repo_url.encode()).decode("utf-8")

    @property
    def wildboar_requires(self):
        return self._wildboar_requires

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def download_url(self):
        return self._bundle_url

    def get_bundles(self):
        return self._bundles

    def _refresh(self, timeout):
        try:
            json_data = json.load(urlopen(self.repo_url, timeout=timeout))
        except json.JSONDecodeError as e:
            raise RuntimeError("Cannot parse the repository") from e

        self._wildboar_requires = json_data["wildboar_requires"]
        self._name = _validate_repository_name(json_data["name"])
        self._version = _validate_version(
            json_data["version"], max_version=JSONRepository.supported_version
        )
        if parse_version(self.wildboar_requires) > parse_version(wildboar_version):
            raise ValueError(
                "The repository requires wildboar >=%s, got %s"
                % (self.wildboar_requires, wildboar_version),
            )
        self._bundle_url = _validate_url(json_data["bundle_url"])
        bundles = {}
        for bundle_json in json_data["bundles"]:
            key = _validate_bundle_key(bundle_json["key"])
            if key in bundles:
                warnings.warn("duplicate dataset %s (ignoring)" % key, UserWarning)

            version = _validate_version(bundle_json["version"])
            tag = bundle_json.get("tag")
            if tag is not None:
                tag = _validate_bundle_key(tag)

            name = bundle_json.get("name")
            if name is None:
                raise ValueError("bundle %s require 'name', but got None" % key)

            arrays = bundle_json.get("arrays")
            if arrays is not None:
                if not isinstance(arrays, list):
                    warnings.warn(
                        "'arrays' should be list, not %r. Using defaults."
                        % type(arrays).__qualname__,
                        UserWarning,
                    )
                    arrays = None

            description = bundle_json.get("description")
            if description is not None:
                if not isinstance(description, str):
                    warnings.warn(
                        "'description' should be str, not %r. Using defaults."
                        % type(description).__qualname__,
                        UserWarning,
                    )
                    description = None

            collections = bundle_json.get("collections")
            if collections is not None:
                collections = _validate_collections(collections)

            bundles[bundle_json["key"]] = NpBundle(
                key=key,
                version=version,
                tag=tag,
                name=name,
                description=description,
                collections=collections,
                arrays=arrays,
            )

        self._bundles = bundles


class RepositoryCollection:
    def __init__(self):
        self.pending_repositories = set()
        self.repositories = set()
        self._initialized = False

    def __getitem__(self, key):
        if not self._initialized:
            self._initialized = True
            self.refresh(timeout=3)

        repository = next(
            (repository for repository in self.repositories if repository.name == key),
            None,
        )

        if repository is None:
            if self.pending_repositories:
                raise ValueError(
                    "The repository %s does not exist, but %d repositories have not "
                    "been refreshed yet." % (key, len(self.pending_repositories))
                )
            else:
                raise ValueError("The repository %s does not exist." % key)

        return repository

    def __delitem__(self, key):
        self.repositories = set(
            repository for repository in self.repositories if repository.name != key
        )

    def __contains__(self, item):
        return any(
            repository for repository in self.repositories if repository.name == item
        )

    def __iter__(self):
        return iter(self.repositories)

    def __len__(self):
        return len(self.repositories)

    def load_repository(self, repository, cache_dir=None):
        if cache_dir and not repository.active:
            cache_dir = os.path.join(cache_dir, repository.identifier)
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir, exist_ok=True)
            repo_path = os.path.join(cache_dir, "repo")
            if os.path.exists(repo_path):
                with open(repo_path, "rb") as f:
                    return pickle.load(f)

        return None

    def save_repository(self, repository, cache_dir=None):
        if cache_dir and repository.active:
            cache_dir = os.path.join(cache_dir, repository.identifier)
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir, exist_ok=True)
            repo_path = os.path.join(cache_dir, "repo")
            with open(repo_path, "wb") as f:
                pickle.dump(repository, f)

    def refresh(self, repository=None, timeout=None, cache_dir=None):
        if repository is None:
            for repository in self.repositories:
                repository.refresh(timeout)
                if repository.active:
                    self.save_repository(repository, cache_dir=cache_dir)

            for repository in self.pending_repositories:
                repository.refresh(timeout)
                if repository.active:
                    self._check_unique(repository.name)
                    self.repositories.add(repository)
                    self.save_repository(repository, cache_dir=cache_dir)
                else:
                    cached_repository = self.load_repository(
                        repository, cache_dir=cache_dir
                    )
                    if cached_repository:
                        self._check_unique(cached_repository.name)
                        self.repositories.add(cached_repository)

            self.pending_repositories = set(
                repository
                for repository in self.pending_repositories
                if not repository.active
            )
        else:
            repository = self[repository]
            repository.refresh(timeout)
            if repository.active:
                self.save_repository(repository, cache_dir=cache_dir)

    def install(self, repository, refresh=True, timeout=None, cache_dir=None):
        if refresh:
            repository.refresh(timeout)

        if repository.active:
            self._check_unique(repository.name)
            self.repositories.add(repository)
        else:
            cached_repository = self.load_repository(repository, cache_dir=cache_dir)
            if cached_repository:
                self._check_unique(cached_repository)
                self.repositories.add(cached_repository)
            else:
                self.pending_repositories.add(repository)

    def _check_unique(self, key):
        if key in {repo.name for repo in self.repositories}:
            raise ValueError(f"Repository with {key} already exists")


class Bundle(metaclass=ABCMeta):
    """
    Base class for handling dataset bundles.

    Parameters
    ----------
    key : str
        A unique key of the bundle.
    version : str
        The version of the bundle.
    name : str
        Human-readable name of the bundle.
    tag : str, optional
        A bundle tag.
    arrays : list
        The arrays of the dataset.
    description : str
        Description of the bundle.
    collections : dict, optional
        A list of collections.
    """

    def __init__(
        self,
        *,
        key,
        version,
        name,
        tag=None,
        arrays=None,
        description=None,
        collections=None,
    ):
        self.key = key
        self.version = version
        self.name = name
        self.description = description
        self.collections = collections
        self.tag = tag or DEFAULT_TAG
        self.arrays = arrays or ["x", "y"]

    def get_filename(self, version=None, tag=None, ext=None):
        """
        Get the cache name of the bundle.

        Parameters
        ----------
        version : str, optional
            The bundle version.
        tag : str, optional
            The tag.
        ext : str, optional
            The extension of the file.

        Returns
        -------
        str
            The filename.
        """
        filename = "%s-v%s" % (self.key, version or self.version)
        if tag:
            filename += "-%s" % tag
        if ext:
            filename += ext
        return filename

    def get_collection(self, collection):
        """
        Get a dataset collection.

        Parameters
        ----------
        collection : str, optional
            The name of the collection.

        Returns
        -------
        list
            List of datasets in the collection.
        """
        if self.collections is None:
            raise ValueError("The collection %s cannot be found." % collection)
        else:
            c = self.collections.get(collection)
            if c is None:
                raise ValueError("The collection %s cannot be found." % collection)
            return c

    def list(self, archive, collection=None):
        """
        List all datasets in this bundle.

        Parameters
        ----------
        archive : ZipFile
            The bundle file.
        collection : str, optional
            The collection name.

        Returns
        -------
        list
            A sorted list of datasets in the bundle.
        """
        names = []
        if collection is not None:
            sample = self.get_collection(collection)
        else:
            sample = None
        for f in archive.filelist:
            path, ext = os.path.splitext(f.filename)
            if self._is_dataset(path, ext):
                filename = os.path.basename(path)
                filename = re.sub("_(TRAIN|TEST)", "", filename)
                if sample is None or filename in sample:
                    names.append(filename)

        return sorted(set(names))

    def load(  # noqa: PLR0912
        self,
        name,
        archive,
    ):
        """
        Load a dataset from the bundle.

        Parameters
        ----------
        name : str
            Name of the dataset.
        archive : ZipFile
            The zip-file bundle.

        Returns
        -------
        x : ndarray
            Data samples.
        y : ndarray
            Data labels.
        n_training_samples : int
            Number of samples that are for training. The value is <= x.shape[0].
        extras : dict, optional
            Extra numpy arrays.
        """
        datasets = []
        for dataset in map(_Dataset, archive.filelist):
            if dataset.filename == name and self._is_dataset(dataset.path, dataset.ext):
                datasets.append(dataset)

        if not datasets:
            raise ValueError("The dataset %s cannot be found." % name)

        train_parts = [
            self._load_array(archive, dataset.file)
            for dataset in datasets
            if dataset.part == "train"
        ]
        test_parts = [
            self._load_array(archive, dataset.file)
            for dataset in datasets
            if dataset.part == "test"
        ]

        data = {}
        for array in self.arrays:
            data[array] = np.concatenate(
                [train_part[array] for train_part in train_parts], axis=0
            )

        sizes = [data[array].shape[0] for array in self.arrays]
        if max(sizes) != min(sizes):
            raise ValueError(
                "Arrays %s must have the same number of samples." % self.arrays
            )

        n_train_samples = sizes[0]
        if test_parts:
            for array in self.arrays:
                train = data[array]
                test = []
                for test_part in test_parts:
                    test.append(test_part[array])
                test = np.concatenate(test, axis=0)

                if test.ndim != train.ndim:
                    raise ValueError(
                        "The training (%d) and testing (%d) parts have incompatible "
                        "rank" % (test.ndim, train.ndim)
                    )

                if test.ndim > 2 and train.shape[1] != test.shape[1]:
                    raise ValueError(
                        "The training (%d) and testing (%d) parts have incompatible "
                        "number dimensions" % (train.shape[1], test.shape[1])
                    )

                if test.ndim == 1:
                    data[array] = np.concatenate([train, test], axis=0)
                else:
                    if test.ndim == 2:
                        shape = (
                            train.shape[0] + test.shape[0],
                            max(test.shape[-1], train.shape[-1]),
                        )
                    elif test.ndim == 3:
                        shape = (
                            train.shape[0] + test.shape[0],
                            test.shape[1],
                            max(test.shape[-1], train.shape[-1]),
                        )
                    else:
                        raise ValueError(
                            "Invalid dataset rank %d, expected <= 3" % test.ndim
                        )

                    merge = np.full(shape, fill_value=EOS, dtype=np.double)
                    merge[: train.shape[0], ..., : train.shape[-1]] = train
                    merge[train.shape[0] :, ..., : test.shape[-1]] = test
                    data[array] = merge

        x = data.pop("x")
        y = data.pop("y")
        return x, y, n_train_samples, data

    @abstractmethod
    def _is_dataset(self, file_name, ext):
        """
        Check if a path and extension is to be considered a dataset.

        The check should be simple and only consider the filename and or
        extension of the file. Validation of the file should be deferred to
        `_load_array`

        Parameters
        ----------
        file_name : str
            Name of the dataset file.

        ext : str
            Extension of the dataset file.

        Returns
        -------
        bool
            True if it is a dataset.
        """
        pass

    @abstractmethod
    def _load_array(self, archive, file):
        """
        Load the file inside the archive and convert to a numpy array.

        Parameters
        ----------
        archive : ZipFile
            The zip-archive in which the file reside.
        file : str
            Path to the file inside the zip-archive.

        Returns
        -------
        ndarray
            The dataset converted to a ndarray.
        """
        pass


class NpBundle(Bundle):
    """Bundle of numpy binary files."""

    def _is_dataset(self, file_name, ext):
        return ext in [".npy", ".npz"]

    def _load_array(self, archive, file):
        data = np.load(archive.open(file))

        if isinstance(data, np.ndarray):
            if "x" in self.arrays and "y" in self.arrays:
                return {"x": data[:, :-1], "y": data[:, -1]}
            elif "x" in self.arrays:
                return {"x": data}
            else:
                raise ValueError("Can't infer arrays to export.")
        else:
            return data


class _Dataset:
    def __init__(self, zip_info):
        self.file = zip_info
        self.path, self.ext = os.path.splitext(zip_info.filename)
        self.filename = os.path.basename(self.path)
        if "_TRAIN" in self.filename:
            self.part = "train"
            self.filename = self.filename.replace("_TRAIN", "")
        elif "_TEST" in self.filename:
            self.part = "test"
            self.filename = self.filename.replace("_TEST", "")
        else:
            self.part = "train"


def _sha1(file, buf_size=65536):
    sha1 = hashlib.sha1()
    with open(file, "rb") as f:
        while True:
            data = f.read(buf_size)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()



# cython: language_level=3
# cython: boundscheck=False
# cython: cdivision=True
# cython: wraparound=False
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

from libc.math cimport NAN
from libc.stdlib cimport free, malloc
from libc.string cimport memcpy
from numpy cimport uint32_t

from ..distance._cdistance cimport MetricList
from ..utils cimport TSArray
from ..utils._misc cimport to_ndarray_double
from ..utils._rand cimport RAND_R_MAX, RandomSampler, rand_int

from ..distance._distance import _METRICS

from ._attr_gen cimport Attribute, AttributeGenerator


cdef struct TransientPivot:
    Py_ssize_t sample
    Py_ssize_t metric

cdef struct PersitentPivot:
    double *data
    Py_ssize_t length
    Py_ssize_t metric

cdef class PivotAttributeGenerator(AttributeGenerator):
    cdef Py_ssize_t n_pivots
    cdef RandomSampler sampler
    cdef MetricList metrics

    def __cinit__(self, Py_ssize_t n_pivots, list metrics, RandomSampler sampler):
        self.n_pivots = n_pivots
        self.metrics = MetricList(metrics)
        self.sampler = sampler

    def __reduce__(self):
        return self.__class__, (self.n_pivots, self.metrics.py_list, self.sampler)

    cdef int reset(self, TSArray X) noexcept nogil:
        cdef Py_ssize_t i
        for i in range(self.metrics.size):
            self.metrics.reset(i, X, X)
        return 0

    cdef Py_ssize_t get_n_attributess(self, TSArray X) noexcept nogil:
        return self.n_pivots

    cdef Py_ssize_t get_n_outputs(self, TSArray X) noexcept nogil:
        return self.get_n_attributess(X)

    cdef Py_ssize_t next_attribute(
            self,
            Py_ssize_t attribute_id,
            TSArray X,
            Py_ssize_t *samples,
            Py_ssize_t n_samples,
            Attribute *transient,
            uint32_t *seed,
    ) noexcept nogil:
        cdef TransientPivot *pivot = <TransientPivot*> malloc(sizeof(TransientPivot))
        pivot.sample = samples[rand_int(0, n_samples, seed)]
        pivot.metric = self.sampler.rand_int(seed)
        transient.dim = rand_int(0, X.shape[1], seed)
        transient.attribute = pivot
        return 0

    cdef Py_ssize_t free_transient(self, Attribute *attribute) noexcept nogil:
        if attribute.attribute != NULL:
            free(attribute.attribute)
            attribute.attribute = NULL
        return 0

    cdef Py_ssize_t free_persistent(self, Attribute *attribute) noexcept nogil:
        cdef PersitentPivot *pivot
        if attribute.attribute != NULL:
            pivot = <PersitentPivot*> attribute.attribute
            free(pivot.data)
            free(attribute.attribute)
            attribute.attribute = NULL
        return 0

    cdef Py_ssize_t init_persistent(
            self,
            TSArray X,
            Attribute *transient,
            Attribute *persistent
    ) noexcept nogil:
        cdef TransientPivot *pivot = <TransientPivot*> transient.attribute
        cdef PersitentPivot *persistent_pivot = <PersitentPivot*> malloc(
            sizeof(PersitentPivot)
        )

        persistent_pivot.data = <double*> malloc(sizeof(double) * X.shape[2])
        persistent_pivot.metric = pivot.metric
        persistent_pivot.length = X.shape[2]
        memcpy(
            persistent_pivot.data,
            &X[pivot.sample, transient.dim, 0],
            sizeof(double) * X.shape[2]
        )

        persistent.dim = transient.dim
        persistent.attribute = persistent_pivot
        return 0

    cdef double transient_value(
            self,
            Attribute *attribute,
            TSArray X,
            Py_ssize_t sample
    ) noexcept nogil:
        cdef TransientPivot* pivot = <TransientPivot*>attribute.attribute
        return self.metrics.distance(
            pivot.metric, X, sample, X, pivot.sample, attribute.dim
        )

    cdef double persistent_value(
            self,
            Attribute *attribute,
            TSArray X,
            Py_ssize_t sample
    ) noexcept nogil:
        cdef PersitentPivot* pivot = <PersitentPivot*> attribute.attribute
        return self.metrics._distance(
            pivot.metric,
            &X[sample, attribute.dim, 0],
            X.shape[2],
            pivot.data,
            X.shape[2],
        )

    cdef Py_ssize_t transient_fill(
            self,
            Attribute *attribute,
            TSArray X,
            Py_ssize_t sample,
            double[:, :] out,
            Py_ssize_t out_sample,
            Py_ssize_t out_attribute,
    ) noexcept nogil:
        out[out_sample, out_attribute] = self.transient_value(attribute, X, sample)
        return 0

    cdef Py_ssize_t persistent_fill(
            self,
            Attribute *attribute,
            TSArray X,
            Py_ssize_t sample,
            double[:, :] out,
            Py_ssize_t out_sample,
            Py_ssize_t out_attribute,
    ) noexcept nogil:
        out[out_sample, out_attribute] = self.persistent_value(attribute, X, sample)
        return 0

    cdef object persistent_to_object(self, Attribute *attribute):
        cdef PersitentPivot *pivot = <PersitentPivot*> attribute.attribute
        return attribute.dim, (
            pivot.metric, to_ndarray_double(pivot.data, pivot.length)
        )

    cdef Py_ssize_t persistent_from_object(self, object object, Attribute *attribute):
        dim, (metric, arr) = object
        cdef PersitentPivot *pivot = <PersitentPivot*> malloc(sizeof(PersitentPivot))
        cdef double *data = <double*> malloc(sizeof(double) * arr.size)
        cdef Py_ssize_t i
        for i in range(arr.size):
            data[i] = arr[i]
        pivot.data = data
        pivot.length = arr.size
        pivot.metric = metric

        attribute.dim = dim
        attribute.attribute = pivot
        return 0



"""Methods for model selection."""
# Authors: Isak Samsten
# License: BSD 3 clause

from ._cv import RepeatedOutlierSplit
from ._outlier import outlier_train_test_split

__all__ = [
    "RepeatedOutlierSplit",
    "outlier_train_test_split",
]



# Authors: Isak Samsten
# License: BSD 3 clause

import math
import numbers
import warnings
from abc import ABCMeta, abstractmethod
from collections import namedtuple

import numpy as np
from sklearn.metrics import check_scoring
from sklearn.metrics._scorer import _check_multimetric_scoring, _MultimetricScorer
from sklearn.model_selection._validation import _aggregate_score_dicts
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.validation import check_is_fitted, check_random_state, check_scalar

from ..base import BaseEstimator, ExplainerMixin
from ..distance import pairwise_subsequence_distance
from ..transform._sax import SAX
from ..transform._shapelet import RandomShapeletTransform, ShapeletMixin
from ..utils.validation import check_array, check_option

try:
    from matplotlib.axes import Axes
    from matplotlib.cm import ScalarMappable, get_cmap
    from matplotlib.pylab import subplots

    from ..utils.plot import MidpointNormalize, plot_frequency_domain, plot_time_domain
except ModuleNotFoundError as e:
    from ..utils import DependencyMissing

    matplotlib_missing = DependencyMissing(e, package="matplotlib")
    get_cmap = matplotlib_missing
    ScalarMappable = matplotlib_missing
    subplots = matplotlib_missing
    MidpointNormalize = matplotlib_missing
    plot_time_domain = matplotlib_missing
    Axes = matplotlib_missing

Importance = namedtuple("Importance", ["mean", "std", "full"])


def plot_importances(importances, ax=None, labels=None):
    """
    Plot the importances as a boxplot.

    Parameters
    ----------
    importances : Importance or dict
        The importances.
    ax : Axes, optional
        The axes to plot. If importances is dict, ax must contain at least
        len(importances) Axes objects.
    labels : array-like, optional
        The labels for the importances.

    Returns
    -------
    Axes
        The plotted Axes.
    """
    n_importances = len(importances) if isinstance(importances, dict) else 1
    if ax is None:
        _, ax = subplots(ncols=n_importances)

    if isinstance(importances, dict):
        for i, (title, importance) in enumerate(importances.items()):
            _plot_single_importance(importance, ax=ax[i], labels=labels)
            ax[i].set_title(title)
    else:
        _plot_single_importance(importances, ax=ax, labels=labels)

    return ax


def _plot_single_importance(importance, ax, labels):
    order = np.argsort(importance.mean)
    if labels is not None:
        labels = np.array(labels)

    bplot = ax.boxplot(
        importance.full[order].T,
        sym="",
        vert=False,
        patch_artist=True,
        labels=order if labels is None else labels[order],
        medianprops=dict(color="black"),
    )

    cmap = get_cmap("Dark2", importance.mean.size)
    for i, patch in enumerate(bplot["boxes"]):
        patch.set_facecolor(cmap(i))


def _intervals(n, n_intervals):
    for i in range(n_intervals):
        length = n // n_intervals
        start = i * length + min(i % n_intervals, n % n_intervals)
        if i % n_intervals < n % n_intervals:
            length += 1
        yield start, start + length


def _unpack_scores(orig_score, perm_score):
    importances = orig_score - perm_score
    return Importance(
        mean=np.mean(importances, axis=1),
        std=np.std(importances, axis=1),
        full=importances,
    )


class PermuteImportance(BaseEstimator, metaclass=ABCMeta):
    _parameter_constraints: dict = {
        "scoring": [None, str, list, dict, callable],
        "n_repeat": [Interval(numbers.Integral, 1, None, closed="left")],
        "random_state": ["random_state"],
    }

    def __init__(self, *, scoring=None, n_repeat=1, random_state=None):
        self.scoring = scoring
        self.n_repeat = n_repeat
        self.random_state = random_state

    @abstractmethod
    def _yield_components(self):
        pass

    def _reset_component(self, X, component):
        pass

    @abstractmethod
    def _permute_component(self, X, component, random_state):
        pass

    def _fit(self, estimator, X, y, random_state, sample_weight=None):  # noqa: PLR0912
        if callable(self.scoring):
            scoring = self.scoring
        elif self.scoring is None or isinstance(self.scoring, str):
            scoring = check_scoring(estimator, self.scoring)
        else:
            scoring_dict = _check_multimetric_scoring(estimator, self.scoring)
            scoring = _MultimetricScorer(**scoring_dict)

        self.components_ = np.array(list(self._yield_components()), dtype=object)
        scores = []
        for component in self.components_:
            self._reset_component(X, component)

            rep_scores = []
            for rep in range(self.n_repeat):
                X_perm = self._permute_component(X, component, random_state)

                if sample_weight is not None:
                    score = scoring(estimator, X_perm, y, sample_weight=sample_weight)
                else:
                    score = scoring(estimator, X_perm, y)

                rep_scores.append(score)

            if isinstance(rep_scores[0], dict):
                scores.append(_aggregate_score_dicts(rep_scores))
            else:
                scores.append(rep_scores)

        if sample_weight is not None:
            self.baseline_score_ = scoring(estimator, X, y, sample_weight=sample_weight)
        else:
            self.baseline_score_ = scoring(estimator, X, y)

        if isinstance(self.baseline_score_, dict):
            self.importances_ = {
                name: _unpack_scores(
                    self.baseline_score_[name],
                    np.array([scores[i][name] for i in range(len(self.components_))]),
                )
                for name in self.baseline_score_
            }
        else:
            self.importances_ = _unpack_scores(self.baseline_score_, np.array(scores))


class IntervalImportance(ExplainerMixin, PermuteImportance):
    """
    Interval importance for time series.

    Parameters
    ----------
    scoring : str, list, dict or callable, optional
        The scoring function. By default the estimators score function is used.
    n_repeat : int, optional
        The number of repeated permutations.
    n_intervals : str, optional
        The number of intervals.

        - if "sqrt", the number of intervals is the square root of n_timestep.
        - if "log2", the number of intervals is the log2 of n_timestep.
        - if int, exact number of intervals.
    window : int, optional
        The window size. If specicied, n_intervals is ignored and the number of
        intervals is computed such that each interval is (at least) of size window.
    random_state : int or RandomState
        - If `int`, `random_state` is the seed used by the random number generator
        - If `RandomState` instance, `random_state` is the random number generator
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    Attributes
    ----------
    importances_ : dict or Importance
        The importance scores for each interval. If dict, one value per scoring
        function.

    components_ : ndarray of shape (n_intervals, 2)
        The interval start and end positions.

    """

    _parameter_constraints: dict = {
        **PermuteImportance._parameter_constraints,
        "n_intervals": [
            StrOptions({"sqrt", "log2", "log"}, deprecated={"log"}),
            Interval(numbers.Integral, 1, None, closed="left"),
        ],
        "window": [Interval(numbers.Integral, 1, None, closed="left"), None],
    }

    def __init__(
        self,
        *,
        scoring=None,
        n_repeat=5,
        n_intervals="sqrt",
        window=None,
        random_state=None,
    ):
        super().__init__(scoring=scoring, n_repeat=n_repeat, random_state=random_state)
        self.n_intervals = n_intervals
        self.window = window

    def _yield_components(self):
        if self.window is not None:
            if self.window > self.n_timesteps_in_:
                raise ValueError(
                    f"The window parameter of {type(self).__name__} must be "
                    "<= n_timesteps_in_"
                )

            n_intervals = self.n_timesteps_in_ // self.window
        elif self.n_intervals == "sqrt":
            n_intervals = math.ceil(math.sqrt(self.n_timesteps_in_))
        elif self.n_intervals in {"log", "log2"}:
            if self.n_intervals == "log":
                warnings.warn(
                    "The value 'log' for parameter n_intervals of "
                    f"{type(self).__name__} has been renamed to log2 in 1.2 and will "
                    "be removed in 1.4"
                )

            n_intervals = math.ceil(math.log2(self.n_timesteps_in_))
        elif isinstance(self.n_intervals, numbers.Integral):
            if self.n_intervals > self.n_timesteps_in_:
                raise ValueError(
                    f"The n_intervals parameter of {type(self).__name__} must be "
                    "<= n_timesteps_in_"
                )
            n_intervals = self.n_intervals
        else:
            n_intervals = math.ceil(self.n_timesteps_in_ * self.n_intervals)

        return _intervals(self.n_timesteps_in_, n_intervals)

    def _permute_component(self, X, component, random_state):
        X_perm = X.copy()
        start, end = component
        random_state.shuffle(X_perm[:, start:end])
        return X_perm

    def fit(self, estimator, x, y, sample_weight=None):
        self._validate_params()
        estimator = self._validate_estimator(estimator)
        x, y = self._validate_data(x, y, reset=False, allow_3d=False)
        random_state = check_random_state(self.random_state)
        self._fit(estimator, x, y, random_state, sample_weight=sample_weight)
        return self

    def explain(self, x, y=None):
        check_is_fitted(self)
        x = self._validate_data(x, reset=False)
        importances = np.empty(self.n_timesteps_in_, dtype=float)
        for i, (start, end) in enumerate(self.components_):
            importances[start:end] = self.importances_.mean[i]
        return np.broadcast_to(importances, (x.shape[0], self.n_timesteps_in_))

    def plot(  # noqa: PLR0912
        self,
        x=None,
        y=None,
        *,
        ax=None,
        scoring=None,
        k=None,
        n_samples=100,
        show_grid=True,
    ):
        check_is_fitted(self)
        if isinstance(self.importances_, dict):
            importances = check_option(self.importances_, scoring, "scoring")
        else:
            if scoring is not None and scoring != self.scoring:
                raise ValueError("scoring must be %s, got %r" % (self.scoring, scoring))
            importances = self.importances_

        if x is None:
            return plot_importances(
                importances,
                ax=ax,
                labels=["(%d, %d)" % (start, end) for start, end in self.components_],
            )

        importances = importances.mean
        if k is None:
            k = importances.shape[0]
        elif isinstance(k, numbers.Integral):
            k = check_scalar(
                k,
                "top_k",
                numbers.Integral,
                min_val=1,
                max_val=importances.shape[0],
            )
        elif isinstance(k, numbers.Real):
            k = check_scalar(
                k,
                "top_k",
                numbers.Real,
                min_val=0,
                max_val=1,
                include_boundaries="right",
            )
            k = math.ceil(k * importances.shape[0])
        else:
            raise TypeError(
                "top_k must be int or float, not %s." % type(k).__qualname__
            )

        if ax is None:
            fig, ax = subplots()
        else:
            fig = None

        order = np.argsort(importances)[: -(k + 1) : -1]
        norm = MidpointNormalize(
            # vmin=importances[order[-1]],
            # vmax=importances[order[0]],
            # midpoint=0.0,
            vmin=0.0,
            vmax=1.0,
            midpoint=0.0,
        )
        cmap = get_cmap("coolwarm")

        if show_grid:
            for start, _ in self.components_:
                ax.axvline(
                    start,
                    0,
                    1,
                    color="gray",
                    linestyle="dashed",
                    linewidth=0.5,
                    dashes=(5, 5),
                    zorder=-2,
                )

        for o in order:
            start, end = self.components_[o]
            ax.axvspan(
                start,
                end,
                0,
                1,
                alpha=0.2,
                color=cmap(norm(importances[o])),
                zorder=1,
            )

        xticks = [start for start, _ in self.components_]
        xticks.append(self.components_[-1][1])
        ax.set_xticks(xticks)
        ax.tick_params(axis="x", labelrotation=-70)

        plot_time_domain(x, y=y, n_samples=n_samples, ax=ax)
        mappable = ScalarMappable(cmap=cmap, norm=norm)
        if fig is not None:
            fig.colorbar(mappable, ax=ax)
            return ax
        else:
            return ax, mappable


class FrequencyImportance(ExplainerMixin, PermuteImportance):
    def __init__(self, window=1, scoring=None, n_repeat=1, random_state=None):
        super().__init__(scoring=scoring, n_repeat=n_repeat, random_state=random_state)
        self.window = window

    # TODO: Ensure it has access to X to compute things in terms of it
    def _yield_components(self):
        n_timesteps = self.n_timesteps_in_ // 2
        n_intervals = n_timesteps
        if self.window > n_timesteps:
            raise ValueError(
                f"The window parameter of {type(self).__name__} must be "
                "<= n_timesteps_in_"
            )

        if n_timesteps % 2 != 0:
            n_timesteps -= 1

        n_intervals = n_timesteps // self.window
        for i in range(n_intervals):
            length = n_timesteps // n_intervals
            start = 1 + i * length + min(i % n_intervals, n_timesteps % n_intervals)
            if i % n_intervals < n_timesteps % n_intervals:
                length += 1
            yield start, start + length

    def _permute_component(self, X, component, random_state):
        start, end = component
        X_fft = np.fft.rfft(X, axis=1)
        random_state.shuffle(X_fft[:, start:end])
        return np.fft.irfft(X_fft, X.shape[-1], axis=1)

    def fit(self, estimator, x, y, sample_weight=None):
        # self._validate_params()
        estimator = self._validate_estimator(estimator)
        x, y = self._validate_data(x, y, reset=False, allow_3d=False)
        random_state = check_random_state(self.random_state)
        self._fit(estimator, x, y, random_state, sample_weight=sample_weight)
        return self

    def plot(self, x=None, y=None, ax=None):
        check_is_fitted(self)

        importances = self.importances_.mean
        if x is None:
            return plot_importances(
                importances,
                ax=ax,
                labels=["(%d, %d)" % (start, end) for start, end in self.components_],
            )

        if ax is None:
            fig, ax = subplots()
        else:
            fig = None

        order = np.argsort(importances)[::-1]
        norm = MidpointNormalize(
            vmin=0,  # importances[order[-1]],
            vmax=1,  # importances[order[0]],
            midpoint=0.0,
        )
        cmap = get_cmap("coolwarm")

        for o in order:
            start, end = self.components_[o]
            ax.axvspan(
                start - 0.5,
                end - 0.5,
                0,
                1,
                alpha=0.2,
                color=cmap(norm(importances[o])),
                zorder=1,
            )

        xticks = [start for start, _ in self.components_]
        xticks.append(self.components_[-1][1])
        ax.set_xticks(xticks)
        ax.tick_params(axis="x", labelrotation=-70)
        plot_frequency_domain(x, y, ax=ax)
        mappable = ScalarMappable(cmap=cmap, norm=norm)
        if fig is not None:
            fig.colorbar(mappable, ax=ax)
            return ax
        else:
            return ax, mappable


class AmplitudeImportance(ExplainerMixin, PermuteImportance):
    """
    Compute the importance of equi-probable amplitude intervals.

    The implementation uses
    :class:`transform.SAX` to discretize the time series and then for each bin permute
    the samples along that bin.
    """

    _parameter_constraints: dict = {
        **IntervalImportance._parameter_constraints,
        "binning": [StrOptions({"normal", "uniform"})],
        "n_bins": [Interval(numbers.Integral, 1, None, closed="left")],
    }

    def __init__(
        self,
        scoring=None,
        n_intervals="sqrt",
        window=None,
        binning="normal",
        n_bins=4,
        n_repeat=1,
        random_state=None,
    ):
        super().__init__(
            scoring=scoring,
            n_repeat=n_repeat,
            random_state=random_state,
        )
        self.n_intervals = n_intervals
        self.window = window
        self.binning = binning
        self.n_bins = n_bins

    def _yield_components(self):
        return (b for b in range(self.n_bins))

    def _permute_component(self, X, component, random_state):
        X_perm = np.empty_like(X)
        idx = np.arange(X.shape[0])
        random_state.shuffle(idx)
        for i in range(X.shape[0]):
            j = idx[i]
            xi = self.x_sax_[i, :]
            xj = self.x_sax_[j, :]

            i_indicies = np.where(xi == component)[0]
            j_indicies = np.where(xj == component)[0]

            if i_indicies.size < j_indicies.size:
                j_indicies = random_state.choice(
                    j_indicies, i_indicies.size, replace=False
                )
            elif i_indicies.size > j_indicies.size:
                i_indicies = random_state.choice(
                    i_indicies, j_indicies.size, replace=False
                )

            X_perm[i] = X[i]
            for i_idx, j_idx in zip(i_indicies, j_indicies):
                i_start, i_end = self.sax_.intervals[i_idx]
                j_start, j_end = self.sax_.intervals[j_idx]

                if i_end - i_start < j_end - j_start:
                    j_end -= 1
                elif j_end - j_start < i_end - i_start:
                    i_end -= 1

                X_perm[i, i_start:i_end] = X[j, j_start:j_end]

        return X_perm

    def fit(self, estimator, X, y, sample_weight=None):
        self._validate_params()
        estimator = self._validate_estimator(estimator)
        X, y = self._validate_data(X, y, reset=False, allow_3d=False)
        self.sax_ = SAX(
            n_intervals=self.n_intervals,
            window=self.window,
            binning=self.binning,
            n_bins=self.n_bins,
            estimate=True,
        ).fit(X)
        self.x_sax_ = self.sax_.transform(X)
        random_state = check_random_state(self.random_state)
        self._fit(estimator, X, y, random_state, sample_weight=sample_weight)
        return self

    def explain(self, X, y=None):
        X = self._validate_data(X, reset=False)
        X_sax = self.sax_.transform(X)
        return self.importances_.mean[X_sax]

    def _validate_data_plot(self, x):
        if (
            self.binning == "uniform"
            and not np.allclose(np.max(x, axis=1), 1)
            and not np.allclose(np.min(x, axis=1), 0)
        ):
            raise ValueError(
                "The SAX binning has been estimated from data, and the data to "
                "plot is not min/max scaled. Set preprocess=True to scale the data."
            )

        if (
            self.binning == "normal"
            and not np.allclose(np.mean(x, axis=1), 0)
            and not np.allclose(np.std(x, axis=1), 1)
        ):
            raise ValueError(
                "The SAX binning has been estimated from data, and the data to "
                "plot is not standardized. Set preprocess=True to scale the data."
            )

    def plot(  # noqa: PLR0915, PLR0912
        self,
        x=None,
        y=None,
        *,
        ax=None,
        n_samples=100,
        scoring=None,
        preprocess=True,
        k=None,
        show_bins=False,
        show_grid=True,
    ):
        """
        Plot the importances.

        If x is given, the importances are plotted over the samples optionally
        labeling each sample using the supplied labels. If x is not give, the
        importances are plotted as one or more boxplots.

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timesteps), optional
            The samples.
        y : array-like of shape (n_samples, ), optional
            The labels.
        ax : Axes, optional
            Axes to plot. If ax is set, x is None and scoring is None, the number of
            axes must be the same as the number of scorers.
        n_samples : int or float, optional
            The number of samples to plot, set to `None` to plot all.
        scoring : str, optional
            The scoring to plot if multiple scorers were used when fitting.
        preprocess : bool, optional
            Preprocess the time series to align with the bins, ignored if x is not None.
        k : int or float, optional
            The number of top bins to plot, ignored if x is not None.

            - if int, the specified number of bins are shown
            - if float, a fraction of the number of bins are shown.
        show_bins : bool, optional
            Annotate the plot with the index of the bin, ignored if x is not None.
        show_grid : bool, optional
            Annotate the plot with the bin thresholds, ignored if x is not None.

        Returns
        -------
        ax : Axis
            The axis.
        mappable : ScalarMappable, optional
            Return the mappable used to plot the colorbar.
            Only returned if ax is not None and x is not None.
        """
        check_is_fitted(self)
        if x is None:
            if scoring is not None and isinstance(self.importances_, dict):
                importances = check_option(self.importances_, scoring, "scoring")
            else:
                importances = self.importances_

            return plot_importances(importances, ax=ax)

        if isinstance(self.importances_, dict):
            importances = check_option(self.importances_, scoring, "scoring")
        else:
            if scoring is not None and scoring != self.scoring:
                raise ValueError(
                    "scoring must be '%s', got %r" % (self.scoring, scoring)
                )
            importances = self.importances_

        x = check_array(x)
        if preprocess:
            x = self.sax_.binning_.scale(x)
        else:
            self._validate_data_plot(x)

        importances = importances.mean
        if k is None:
            k = importances.shape[0]
        elif isinstance(k, numbers.Integral):
            k = check_scalar(
                k,
                "top_k",
                numbers.Integral,
                min_val=1,
                max_val=importances.shape[0],
            )
        elif isinstance(k, numbers.Real):
            k = check_scalar(
                k,
                "top_k",
                numbers.Real,
                min_val=0,
                max_val=1,
                include_boundaries="right",
            )
            k = math.ceil(k * importances.shape[0])
        else:
            raise TypeError(
                "top_k must be int or float, not %s." % type(k).__qualname__
            )

        if ax is None:
            fig, ax = subplots()
        else:
            fig = None

        order = np.argsort(importances)[: -(k + 1) : -1]
        norm = MidpointNormalize(
            # vmin=importances[order[-1]], vmax=importances[order[0]], midpoint=0.0
            vmin=0,
            vmax=1,
            midpoint=0.0,
        )
        thresholds = self.sax_.binning_.get_thresholds()
        cmap = get_cmap("coolwarm")
        bottom, top = np.min(x), np.max(x)

        if show_grid:
            for threshold in thresholds:
                ax.axhline(
                    threshold,
                    0,
                    1,
                    color="gray",
                    linestyle="dashed",
                    linewidth=0.5,
                    dashes=(5, 5),
                    zorder=-2,
                )

        for i in order:
            if i == 0:
                start = bottom
                end = thresholds[i]
            elif i == importances.shape[0] - 1:
                start = thresholds[i - 1]
                end = top
            else:
                start = thresholds[i - 1]
                end = thresholds[i]

            span = ax.axhspan(
                start,
                end,
                0,
                1,
                alpha=0.3,
                zorder=1,
                color=cmap(norm(importances[i])),
                label="%i" % i,
            )

            if show_bins:
                ax.annotate("%d" % i, xy=(0, 0.5), va="center", xycoords=span)

        plot_time_domain(x, y, ax=ax, n_samples=n_samples)
        ax.set_ylim([np.min(x) - np.std(x), np.max(x) + np.std(x)])
        mappable = ScalarMappable(cmap=cmap, norm=norm)
        if fig is not None:
            fig.colorbar(mappable, ax=ax)
            return ax
        else:
            return ax, mappable


class ShapeletImportance(ExplainerMixin, PermuteImportance):
    """
    Compute the importance of shapelets.

    The importance is given by permuting time series sections with
    the minimum distance to shapelets.

    Parameters
    ----------
    scoring : str, list, dict or callable, optional
        The scoring function. By default the estimators score function is used.
    n_repeat : int, optional
        The number of repeated permutations.
    n_shapelets : int, optional
        The number of shapelets to sample for the explanation.
    min_shapelet_size : float, optional
        The minimum size of shapelets used for explanation.
    max_shapelet_size : float, optional
        The maximum size of shapelets used for explanation.
    metric : str, optional
        The metric.
    metric_params : str, optional
        The metric parameters.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If `int`, `random_state` is the seed used by the
          random number generator.
        - If :class:`numpy.random.RandomState` instance, `random_state` is the
          random number generator.
        - If `None`, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.

    Attributes
    ----------
    components : ndarray
        The shapelets

    """

    _parameter_constraints: dict = {
        **PermuteImportance._parameter_constraints,
        **ShapeletMixin._parameter_constraints,
    }

    def __init__(
        self,
        scoring=None,
        n_repeat=1,
        n_shapelets=10,
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        metric="euclidean",
        metric_params=None,
        random_state=None,
    ):
        super().__init__(
            scoring=scoring,
            n_repeat=n_repeat,
            random_state=random_state,
        )
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params

    def _reset_component(self, X, component):
        _, self._shapelet_idx = pairwise_subsequence_distance(
            component,
            X,
            metric=self.metric,
            metric_params=self.metric_params,
            return_index=True,
        )

    def _yield_components(self):
        for _dim, (_, shapelet) in self.shapelet_transform_.embedding_.attributes:
            yield shapelet

    def _permute_component(self, X, component, random_state):
        X_perm = np.empty_like(X)
        idx = np.arange(X.shape[0])
        random_state.shuffle(idx)
        for i in range(X.shape[0]):
            j = idx[i]

            i_shapelet_idx = self._shapelet_idx[i]
            j_shapelet_idx = self._shapelet_idx[j]
            X_perm[j] = X[j]
            if i != j:
                X_perm[j, j_shapelet_idx : (j_shapelet_idx + component.size)] = X[
                    i, i_shapelet_idx : (i_shapelet_idx + component.size)
                ]

        return X_perm

    def fit(self, estimator, X, y, sample_weight=None):
        self._validate_params()
        self._validate_estimator(estimator)
        X, y = self._validate_data(X, y, reset=False)

        random_state = check_random_state(self.random_state)
        self.shapelet_transform_ = RandomShapeletTransform(
            n_shapelets=self.n_shapelets,
            metric=self.metric,
            min_shapelet_size=self.min_shapelet_size,
            max_shapelet_size=self.max_shapelet_size,
            random_state=random_state.randint(np.iinfo(np.int32).max),
        ).fit(X)

        self._fit(estimator, X, y, random_state, sample_weight=sample_weight)
        return self

    def explain(self, X, y=None, kernel_scale=0.25):
        distances, indices = pairwise_subsequence_distance(
            self.components_,
            X,
            metric=self.metric,
            metric_params=self.metric_params,
            return_index=True,
        )

        weights = (
            self._distance_weight(distances, kernel_scale)
            if kernel_scale is not None
            else np.broadcast_to(1, shape=(X.shape[0], len(self.components_)))
        )
        importances = self.importances_.mean

        if y is None:
            explanation = np.zeros((len(self.components_), X.shape[-1]), dtype=float)
            for j, shapelet in enumerate(self.components_):
                importance = importances[j]
                for i in range(X.shape[0]):
                    index = indices[i, j]
                    weight = weights[i, j]
                    explanation[j, index : (index + shapelet.size)] += (
                        weight * importance
                    )

            return explanation / X.shape[0]
        else:
            labels, inv, counts = np.unique(y, return_inverse=True, return_counts=True)
            explanation = np.zeros(
                (len(self.components_), len(labels), X.shape[-1]), dtype=float
            )
            for j, shapelet in enumerate(self.components_):
                importance = importances[j]
                for i in range(X.shape[0]):
                    index = indices[i, j]
                    weight = weights[i, j]
                    explanation[j, inv[i], index : (index + shapelet.size)] += (
                        1 / counts[inv[i]] * weight * importance
                    )

            return explanation

    def plot(  # noqa: PLR0912
        self, X=None, y=None, k=None, scoring=None, kernel_scale=0.25, ax=None
    ):
        if X is None:
            return plot_importances(
                self.importances_, ax=ax, labels=range(len(self.components_))
            )

        if isinstance(self.importances_, dict):
            importances = check_option(self.importances_, scoring, "scoring")
        else:
            if scoring is not None and scoring != self.scoring:
                raise ValueError(
                    "scoring must be '%s', got %r" % (self.scoring, scoring)
                )
            importances = self.importances_

        importances = importances.mean
        order = np.argsort(importances)[::-1]

        if k is None:
            k = importances.size

        if y is None:
            explanation = self.explain(X, kernel_scale=kernel_scale)
            if ax is None:
                fig, ax = subplots()
            else:
                fig = None

            mappable = ax.pcolormesh(explanation[order[:k]], cmap="coolwarm")
            ax.set_yticks(np.arange(k) + 0.5, order[:k])
            if fig is not None:
                fig.colorbar(mappable, ax=ax)
                return ax
            else:
                return ax, mappable
        else:
            explanation = self.explain(X, y, kernel_scale=kernel_scale)
            distances, index = pairwise_subsequence_distance(
                self.components_,
                X,
                metric=self.metric,
                metric_params=self.metric_params,
                return_index=True,
            )
            cmap = get_cmap("coolwarm")
            norm = MidpointNormalize(
                # vmin=importances[:k].min(), vmax=importances[:k].max(), midpoint=0
                vmin=0.0,
                vmax=1.0,
                midpoint=0.0,
            )
            labels, inv, lbl_count = np.unique(
                y, return_inverse=True, return_counts=True
            )
            if ax is None:
                fig, ax = subplots(
                    nrows=k,
                    ncols=len(labels),
                    figsize=(len(labels) * 4, k * 2.5),
                    sharex=True,
                    sharey=True,
                )

            weights = (
                self._distance_weight(distances, kernel_scale)
                if kernel_scale is not None
                else np.broadcast_to(1, shape=(X.shape[0], len(self.components_)))
            )
            for i in range(len(labels)):
                for j in range(k):
                    plot_time_domain(
                        X[y == labels[i]], n_samples=lbl_count[i], ax=ax[j, i]
                    )

            for i in range(0, X.shape[0]):
                for j in range(k):
                    order_j = order[j]
                    ax[j, inv[i]].plot(
                        np.arange(
                            index[i, order_j],
                            index[i, order_j] + self.components_[order_j].size,
                        ),
                        self.components_[order_j],
                        linewidth=weights[i, order_j],
                        color=cmap(norm(importances[order_j])),
                    )

            for i in range(len(labels)):
                ax[0, i].set_title("Label=%r" % labels[i])

            for i in range(k):
                ax[i, 0].set_ylabel(order[i])

            mappable = ScalarMappable(norm=norm, cmap=cmap)
            if fig is not None:
                fig.colorbar(mappable, orientation="horizontal", ax=ax[-1, :])
                return ax
            else:
                return ax, mappable

    def _distance_weight(self, distances, kernel_scale=0.25):
        kernel_width = [np.sqrt(s.size) * kernel_scale for s in self.components_]
        return np.sqrt(np.exp(-(distances**2) / np.array(kernel_width) ** 2))



# Authors: Isak Samsten
# License: BSD 3 clause

import math

import numpy as np
from sklearn import model_selection
from sklearn.utils import check_random_state
from sklearn.utils._random import sample_without_replacement


def outlier_train_test_split(
    x, y, normal_class, test_size=0.2, anomalies_train_size=0.05, random_state=None
):
    """
    Outlier training and testing split from classification dataset.

    Parameters
    ----------
    x : array-like of shape (n_samples, n_timestep) or (n_samples, n_dim, n_timestep)
        Input data samples.
    y : array-like of shape (n_samples,)
        Input class label.
    normal_class : int
        Class label that should be considered as the normal class.
    test_size : float, optional
        Size of the test set.
    anomalies_train_size : float, optional
        Contamination of anomalies in the training dataset.
    random_state : int or RandomState, optional
        Psudo random state used for stable results.

    Returns
    -------
    x_train : array-like
        Training samples.
    x_test : array-like
        Test samples.
    y_train : array-like
        Training labels (either 1 or -1, where 1 denotes normal and -1 anomalous).
    y_test : array-like
        Test labels (either 1 or -1, where 1 denotes normal and -1 anomalous).

    Examples
    --------
    >>> from wildboar.datasets import load_two_lead_ecg
    >>> x, y = load_two_lead_ecg()
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     x, y, 1, test_size=0.2, anomalies_train_size=0.05
    ... )

    """
    random_state = check_random_state(random_state)
    normal = y == normal_class
    y = y.copy()
    y[normal] = 1
    y[~normal] = -1

    x_normal = x[np.where(y == 1)]
    x_anomalous = x[np.where(y == -1)]
    y_normal = y[np.where(y == 1)]
    y_anomalous = y[np.where(y == -1)]

    (xn_train, xn_test, yn_train, yn_test) = model_selection.train_test_split(
        x_normal, y_normal, test_size=test_size, random_state=random_state
    )

    n_sample = min(x_anomalous.shape[0], x_normal.shape[0])
    idx = sample_without_replacement(
        x_anomalous.shape[0], n_sample, random_state=random_state
    )
    n_training_anomalies = math.ceil(xn_train.shape[0] * anomalies_train_size)
    idx = idx[:n_training_anomalies]

    x_anomalous_train = x_anomalous[idx, :]
    y_anomalous_train = y_anomalous[idx]

    x_anomalous_test = np.delete(x_anomalous, idx, axis=0)
    y_anomalous_test = np.delete(y_anomalous, idx)

    x_train = np.vstack([xn_train, x_anomalous_train])
    y_train = np.hstack([yn_train, y_anomalous_train])
    x_test = np.vstack([xn_test, x_anomalous_test])
    y_test = np.hstack([yn_test, y_anomalous_test])
    return x_train, x_test, y_train, y_test
